# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

"""Tests related to parallax correction."""

import datetime
import logging
import math
import os
import unittest.mock
import warnings

import dask.array as da
import dask.config
import numpy as np
import pyorbital.tlefile
import pyresample.kd_tree
import pytest
import xarray as xr
from pyproj import Geod
from pyresample import create_area_def

import satpy.resample
from satpy.enhancements.enhancer import get_enhanced_image

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path
# - caplog
# - request


@pytest.fixture
def fake_tle():
    """Produce fake Two Line Element (TLE) object from pyorbital."""
    return pyorbital.tlefile.Tle(
        "Meteosat-42",
        line1="1 40732U 15034A   22011.84285506  .00000004  00000+0  00000+0 0 9995",
        line2="2 40732   0.2533 325.0106 0000976 118.8734 330.4058  1.00272123 23817")


def _get_fake_areas(center, sizes, resolution, code=4326):  # noqa: D417
    """Get multiple square areas with the same center.

    Returns multiple square areas centered at the same location

    Args:
        center (Tuple[float, float]): Center of all areass
        sizes (List[int]): Sizes of areas
        resolution (float): Resolution of fake area.

    Returns:
        List of areas.
    """
    return [create_area_def(
        "fribullus_xax",
        code,
        units="degrees",
        resolution=resolution,
        center=center,
        shape=(size, size))
        for size in sizes]


def _get_attrs(lat, lon, height=35_000_000):
    """Get attributes for datasets in fake scene."""
    return {
        "orbital_parameters": {
            "satellite_actual_altitude": height,  # in m above surface
            "satellite_actual_longitude": lon,
            "satellite_actual_latitude": lat},
        "units": "m"  # does not apply to orbital parameters, I think!
        }


class TestForwardParallax:
    """Test the forward parallax function with various inputs."""

    def test_get_parallax_corrected_lonlats_ssp(self):
        """Test that at SSP, parallax correction does nothing."""
        from satpy.modifiers.parallax import get_parallax_corrected_lonlats
        sat_lat = sat_lon = lon = lat = 0.
        height = 5000.  # m
        sat_alt = 30_000_000.  # m
        corr_lon, corr_lat = get_parallax_corrected_lonlats(
            sat_lon, sat_lat, sat_alt, lon, lat, height)
        assert corr_lon == corr_lat == 0

    def test_get_parallax_corrected_lonlats_clearsky(self):
        """Test parallax correction for clearsky case (returns NaN)."""
        from satpy.modifiers.parallax import get_parallax_corrected_lonlats
        sat_lat = sat_lon = 0
        lat = np.linspace(-20, 20, 25).reshape(5, 5)
        lon = np.linspace(-20, 20, 25).reshape(5, 5).T
        height = np.full((5, 5), np.nan)  # no CTH --> clearsky
        sat_alt = 35_000_000.  # m above surface
        (corr_lon, corr_lat) = get_parallax_corrected_lonlats(
            sat_lon, sat_lat, sat_alt, lon, lat, height)
        # clearsky becomes NaN
        assert np.isnan(corr_lon).all()
        assert np.isnan(corr_lat).all()

    @pytest.mark.parametrize(("lat", "lon"), [(0, 0), (0, 40), (0, 179.9)])
    @pytest.mark.parametrize("resolution", [0.01, 0.5, 10])
    def test_get_parallax_corrected_lonlats_cloudy_ssp(self, lat, lon, resolution):
        """Test parallax correction for fully cloudy scene at SSP."""
        from satpy.modifiers.parallax import get_parallax_corrected_lonlats

        N = 5
        lats = np.linspace(lat-N*resolution, lat+N*resolution, 25).reshape(N, N)
        lons = np.linspace(lon-N*resolution, lon+N*resolution, 25).reshape(N, N).T
        height = np.full((N, N), 10_000)  # constant high clouds at 10 km
        sat_alt = 35_000_000.  # satellite at 35 Mm
        (corr_lon, corr_lat) = get_parallax_corrected_lonlats(
            lon, lat, sat_alt, lons, lats, height)
        # confirm movements behave as expected
        geod = Geod(ellps="sphere")
        # need to use np.tile here as geod.inv doesn't seem to broadcast (not
        # when turning lon/lat in arrays of size (1, 1) either)
        corr_dist = geod.inv(np.tile(lon, [N, N]), np.tile(lat, [N, N]), corr_lon, corr_lat)[2]
        corr_delta = geod.inv(corr_lon, corr_lat, lons, lats)[2]
        uncorr_dist = geod.inv(np.tile(lon, [N, N]), np.tile(lat, [N, N]), lons, lats)[2]
        # should be equal at SSP and nowhere else
        np.testing.assert_allclose(corr_delta[2, 2], 0, atol=1e-9)
        assert np.isclose(corr_delta, 0, atol=1e-9).sum() == 1
        # should always get closer to SSP
        assert (uncorr_dist - corr_dist >= -1e-8).all()
        # should be larger the further we get from SSP
        assert (np.diff(corr_delta[N//2, :N//2+1]) < 0).all()
        assert (np.diff(corr_delta[N//2, N//2:]) > 0).all()
        assert (np.diff(corr_delta[N//2:, N//2]) > 0).all()
        assert (np.diff(corr_delta[:N//2+1, N//2]) < 0).all()
        assert (np.diff(np.diag(corr_delta)[:N//2+1]) < 0).all()
        assert (np.diff(np.diag(corr_delta)[N//2:]) > 0).all()

    def test_get_parallax_corrected_lonlats_cloudy_slant(self):
        """Test parallax correction for fully cloudy scene (not SSP)."""
        from satpy.modifiers.parallax import get_parallax_corrected_lonlats
        sat_lat = sat_lon = 0
        lat = np.linspace(-20, 20, 25).reshape(5, 5)
        lon = np.linspace(-20, 20, 25).reshape(5, 5).T
        height = np.full((5, 5), 10_000)  # constant high clouds at 10 km
        sat_alt = 35_000_000.  # satellite at 35 Mm
        (corr_lon, corr_lat) = get_parallax_corrected_lonlats(
            sat_lon, sat_lat, sat_alt, lon, lat, height)
        # reference value from Simon Proud
        np.testing.assert_allclose(
            corr_lat[4, 4], 19.955, rtol=5e-4)
        np.testing.assert_allclose(
            corr_lon[4, 4], 19.960, rtol=5e-4)

    def test_get_parallax_corrected_lonlats_mixed(self):
        """Test parallax correction for mixed cloudy case."""
        from satpy.modifiers.parallax import get_parallax_corrected_lonlats

        sat_lon = sat_lat = 0
        sat_alt = 35_785_831.0  # m
        lon = da.array([[-20, -10, 0, 10, 20]]*5)
        lat = da.array([[-20, -10, 0, 10, 20]]*5).T
        alt = da.array([
            [np.nan, np.nan, 5000., 6000., np.nan],
            [np.nan, 6000., 7000., 7000., 7000.],
            [np.nan, 7000., 8000., 9000., np.nan],
            [np.nan, 7000., 7000., 7000., np.nan],
            [np.nan, 4000., 3000., np.nan, np.nan]])
        (corrected_lon, corrected_lat) = get_parallax_corrected_lonlats(
            sat_lon, sat_lat, sat_alt, lon, lat, alt)
        assert corrected_lon.shape == lon.shape
        assert corrected_lat.shape == lat.shape
        # lon/lat should be nan for clear-sky pixels
        assert np.isnan(corrected_lon[np.isnan(alt)]).all()
        assert np.isnan(corrected_lat[np.isnan(alt)]).all()
        # otherwise no nans
        assert np.isfinite(corrected_lon[~np.isnan(alt)]).all()
        assert np.isfinite(corrected_lat[~np.isnan(alt)]).all()

    def test_get_parallax_corrected_lonlats_horizon(self):
        """Test that exception is raised if satellites exactly at the horizon.

        Test the rather unlikely case of a satellite elevation of exactly 0
        """
        from satpy.modifiers.parallax import get_parallax_corrected_lonlats
        sat_lat = sat_lon = lon = lat = 0.
        height = 5000.
        sat_alt = 30_000_000.
        with unittest.mock.patch("satpy.modifiers.parallax.get_observer_look") as smpg:
            smpg.return_value = (0, 0)
            with pytest.raises(NotImplementedError):
                get_parallax_corrected_lonlats(sat_lon, sat_lat, sat_alt, lon, lat, height)

    def test_get_surface_parallax_displacement(self):
        """Test surface parallax displacement."""
        from satpy.modifiers.parallax import get_surface_parallax_displacement

        val = get_surface_parallax_displacement(
                0, 0, 36_000_000, 0, 10, 10_000)
        np.testing.assert_allclose(val, 2075.6863144582703)


@pytest.mark.filterwarnings("ignore:Overlap checking not implemented:UserWarning")
class TestParallaxCorrectionClass:
    """Test that the ParallaxCorrection class is behaving sensibly."""

    @pytest.mark.parametrize("center", [(0, 0), (80, -10), (-180, 5)])
    @pytest.mark.parametrize("sizes", [[5, 9]])
    @pytest.mark.parametrize("resolution", [0.05, 1, 10])
    def test_init_parallaxcorrection(self, center, sizes, resolution):
        """Test that ParallaxCorrection class can be instantiated."""
        from satpy.modifiers.parallax import ParallaxCorrection
        fake_area = _get_fake_areas(center, sizes, resolution)[0]
        pc = ParallaxCorrection(fake_area)
        assert pc.base_area == fake_area

    @pytest.mark.parametrize(("sat_pos", "ar_pos"),
                             [((0, 0), (0, 0)), ((0, 0), (40, 0))])
    @pytest.mark.parametrize("resolution", [0.01, 0.5, 10])
    def test_correct_area_clearsky(self, sat_pos, ar_pos, resolution, caplog):
        """Test that ParallaxCorrection doesn't change clearsky geolocation."""
        from satpy.modifiers.parallax import ParallaxCorrection
        from satpy.tests.utils import make_fake_scene
        (sat_lat, sat_lon) = sat_pos
        (ar_lat, ar_lon) = ar_pos
        small = 5
        large = 9
        (fake_area_small, fake_area_large) = _get_fake_areas(
                (ar_lon, ar_lat), [small, large], resolution)
        corrector = ParallaxCorrection(fake_area_small)

        sc = make_fake_scene(
                {"CTH_clear": np.full((large, large), np.nan)},
                daskify=False,
                area=fake_area_large,
            common_attrs=_get_attrs(sat_lat, sat_lon, 35_000_000))

        with caplog.at_level(logging.DEBUG):
            new_area = corrector(sc["CTH_clear"])
        assert "Calculating parallax correction using heights from CTH_clear" in caplog.text
        np.testing.assert_allclose(
                new_area.get_lonlats(),
                fake_area_small.get_lonlats())

    @pytest.mark.parametrize(("lat", "lon"),
                             [(0, 0), (0, 40), (0, 180),
                              (90, 0)])  # relevant for Арктика satellites
    @pytest.mark.parametrize("resolution", [0.01, 0.5, 10])
    def test_correct_area_ssp(self, lat, lon, resolution):
        """Test that ParallaxCorrection doesn't touch SSP."""
        from satpy.modifiers.parallax import ParallaxCorrection
        from satpy.tests.utils import make_fake_scene
        codes = {
                (0, 0): 4326,
                (0, 40): 4326,
                (0, 180): 3575,
                (90, 0): 3575}
        small = 5
        large = 9
        (fake_area_small, fake_area_large) = _get_fake_areas(
                (lon, lat), [small, large], resolution,
                code=codes[(lat, lon)])
        corrector = ParallaxCorrection(fake_area_small)

        sc = make_fake_scene(
                {"CTH_constant": np.full((large, large), 10000)},
                daskify=False,
                area=fake_area_large,
            common_attrs=_get_attrs(lat, lon, 35_000_000))
        new_area = corrector(sc["CTH_constant"])
        assert new_area.shape == fake_area_small.shape
        old_lonlats = fake_area_small.get_lonlats()
        new_lonlats = new_area.get_lonlats()
        if lat != 90:  # don't check SSP longitude if lat=90
            np.testing.assert_allclose(
                    old_lonlats[0][2, 2],
                    new_lonlats[0][2, 2],
                    atol=1e-9)
            np.testing.assert_allclose(
                    old_lonlats[0][2, 2],
                    lon,
                    atol=1e-9)
        np.testing.assert_allclose(
                old_lonlats[1][2, 2],
                new_lonlats[1][2, 2],
                atol=1e-9)
        np.testing.assert_allclose(
                old_lonlats[1][2, 2],
                lat,
                atol=1e-9)

    @pytest.mark.parametrize("daskify", [False, True])
    def test_correct_area_partlycloudy(self, daskify):
        """Test ParallaxCorrection for partly cloudy situation."""
        from satpy.modifiers.parallax import ParallaxCorrection
        from satpy.tests.utils import make_fake_scene
        small = 5
        large = 9
        (fake_area_small, fake_area_large) = _get_fake_areas(
                (0, 50), [small, large], 0.1)
        (fake_area_lons, fake_area_lats) = fake_area_small.get_lonlats()
        corrector = ParallaxCorrection(fake_area_small)

        sc = make_fake_scene(
               {"CTH": np.array([
                    [np.nan, np.nan, 5000., 6000., 7000., 6000., 5000., np.nan, np.nan],
                    [np.nan, 6000., 7000., 7000., 7000., np.nan, np.nan, np.nan, np.nan],
                    [np.nan, 7000., 8000., 9000., np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, 7000., 7000., 7000., np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, 4000., 3000., np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
                    [np.nan, np.nan, 5000., 8000., 8000., 8000., 6000., np.nan, np.nan],
                    [np.nan, 9000., 9000., 9000., 9000., 9000., 9000., 9000., np.nan],
                    [np.nan, 9000., 9000., 9000., 9000., 9000., 9000., 9000., np.nan],
                    [np.nan, 9000., 9000., 9000., 9000., 9000., 9000., 9000., np.nan],
                    ])},
               daskify=daskify,
               area=fake_area_large,
            common_attrs=_get_attrs(0, 0, 40_000_000))
        new_area = corrector(sc["CTH"])
        assert new_area.shape == fake_area_small.shape
        (new_lons, new_lats) = new_area.get_lonlats()
        assert fake_area_lons[3, 4] != new_lons[3, 4]

        np.testing.assert_allclose(
            new_lons,
            np.array([[np.nan, np.nan, 0., 0.1, 0.2],
                      [-0.20077639, -0.10043652, 0., 0.1, 0.2],
                      [-0.20067643, -0.10033821, 0., 0.1, 0.2],
                      [np.nan, np.nan, np.nan, np.nan, np.nan],
                      [-0.20047906, -0.10038274, 0., 0.10038274, 0.20057462]]),
            rtol=1e-5)
        np.testing.assert_allclose(
            new_lats,
            np.array([[np.nan, np.nan, 50.2, 50.2, 50.2],
                      [50.20963341, 50.22331809, 50.1, 50.1, 50.1],
                      [50.09555017, 50.09555005, 50., 50., 50.],
                      [np.nan, np.nan, np.nan, np.nan, np.nan],
                      [49.86771297, 49.90828968, 49.90828963, 49.90828968, 49.88124283]]),
            rtol=1e-6)

    @pytest.mark.parametrize(("res1", "res2"), [(0.08, 0.3), (0.3, 0.08)])
    def test_correct_area_clearsky_different_resolutions(self, res1, res2):
        """Test clearsky correction when areas have different resolutions."""
        from satpy.modifiers.parallax import ParallaxCorrection
        from satpy.tests.utils import make_fake_scene

        # areas with different resolutions, but same coverage

        area1 = create_area_def(
            "fribullus_xax",
            4326,
            units="degrees",
            resolution=res1,
            area_extent=[-1, -1, 1, 1])

        area2 = create_area_def(
            "fribullus_xax",
            4326,
            units="degrees",
            resolution=res2,
            area_extent=[-1, -1, 1, 1])

        with warnings.catch_warnings():
            warnings.simplefilter("error")
            sc = make_fake_scene(
                    {"CTH_clear": np.full(area1.shape, np.nan)},
                    daskify=False,
                    area=area1,
                common_attrs=_get_attrs(0, 0, 35_000_000))

        corrector = ParallaxCorrection(area2)
        new_area = corrector(sc["CTH_clear"])
        np.testing.assert_allclose(
                new_area.get_lonlats(),
                area2.get_lonlats())

    @pytest.mark.xfail(reason="awaiting pyresample fixes")
    def test_correct_area_cloudy_no_overlap(self, ):
        """Test cloudy correction when areas have no overlap."""
        from satpy.modifiers.parallax import MissingHeightError, ParallaxCorrection
        from satpy.tests.utils import make_fake_scene
        areas_00 = _get_fake_areas((0, 40), [5, 9], 0.1)
        areas_shift = _get_fake_areas((90, 20), [5, 9], 0.1)
        fake_area_small = areas_00[0]
        fake_area_large = areas_shift[1]

        sc = make_fake_scene(
                {"CTH_constant": np.full((9, 9), 10000)},
                daskify=False,
                area=fake_area_large,
            common_attrs=_get_attrs(0, 0, 35_000_000))

        corrector = ParallaxCorrection(fake_area_small)
        with pytest.raises(MissingHeightError):
            corrector(sc["CTH_constant"])

    @pytest.mark.xfail(reason="awaiting pyresample fixes")
    def test_correct_area_cloudy_partly_shifted(self, ):
        """Test cloudy correction when areas overlap only partly."""
        from satpy.modifiers.parallax import IncompleteHeightWarning, ParallaxCorrection
        from satpy.tests.utils import make_fake_scene
        areas_00 = _get_fake_areas((0, 40), [5, 9], 0.1)
        areas_shift = _get_fake_areas((0.5, 40), [5, 9], 0.1)
        fake_area_small = areas_00[0]
        fake_area_large = areas_shift[1]

        sc = make_fake_scene(
                {"CTH_constant": np.full((9, 9), 10000)},
                daskify=False,
                area=fake_area_large,
            common_attrs=_get_attrs(0, 0, 35_000_000))

        corrector = ParallaxCorrection(fake_area_small)

        with pytest.warns(IncompleteHeightWarning):
            new_area = corrector(sc["CTH_constant"])
        assert new_area.shape == fake_area_small.shape

    def test_correct_area_cloudy_same_area(self, ):
        """Test cloudy correction when areas are the same."""
        from satpy.modifiers.parallax import ParallaxCorrection
        from satpy.tests.utils import make_fake_scene
        area = _get_fake_areas((0, 0), [9], 0.1)[0]

        sc = make_fake_scene(
                {"CTH_constant": np.full((9, 9), 10000)},
                daskify=False,
                area=area,
            common_attrs=_get_attrs(0, 0, 35_000_000))

        corrector = ParallaxCorrection(area)
        corrector(sc["CTH_constant"])

    def test_correct_area_no_orbital_parameters(self, caplog, fake_tle):
        """Test ParallaxCorrection when CTH has no orbital parameters.

        Some CTH products, such as NWCSAF-GEO, do not include information
        on satellite location directly.  Rather, they include platform name,
        sensor, start time, and end time, that we have to use instead.
        """
        from satpy.modifiers.parallax import ParallaxCorrection
        from satpy.tests.utils import make_fake_scene
        small = 5
        large = 9
        (fake_area_small, fake_area_large) = _get_fake_areas(
                (0, 0), [small, large], 0.05)
        corrector = ParallaxCorrection(fake_area_small)

        sc = make_fake_scene(
                {"CTH_clear": np.full((large, large), np.nan)},
                daskify=False,
                area=fake_area_large,
                common_attrs={
                    "platform_name": "Meteosat-42",
                    "sensor": "irives",
                    "start_time": datetime.datetime(3021, 11, 30, 12, 24, 17),
                    "end_time": datetime.datetime(3021, 11, 30, 12, 27, 22)})
        with unittest.mock.patch("pyorbital.tlefile.read") as plr:
            plr.return_value = fake_tle
            with caplog.at_level(logging.WARNING):
                new_area = corrector(sc["CTH_clear"])
        assert "Orbital parameters missing from metadata." in caplog.text
        np.testing.assert_allclose(
                new_area.get_lonlats(),
                fake_area_small.get_lonlats())


@pytest.mark.filterwarnings("ignore:Overlap checking not implemented:UserWarning")
class TestParallaxCorrectionModifier:
    """Test that the parallax correction modifier works correctly."""

    def test_parallax_modifier_interface(self):
        """Test the modifier interface."""
        from satpy.modifiers.parallax import ParallaxCorrectionModifier
        (area_small, area_large) = _get_fake_areas((0, 0), [5, 9], 0.1)
        fake_bt = xr.DataArray(
                np.linspace(220, 230, 25).reshape(5, 5),
                dims=("y", "x"),
            attrs={"area": area_small, **_get_attrs(0, 0, 35_000_000)})
        cth_clear = xr.DataArray(
                np.full((9, 9), np.nan),
                dims=("y", "x"),
            attrs={"area": area_large, **_get_attrs(0, 0, 35_000_000)})
        modif = ParallaxCorrectionModifier(
                name="parallax_corrected_dataset",
                prerequisites=[fake_bt, cth_clear],
                optional_prerequisites=[],
                cth_radius_of_influence=48_000,
                dataset_radius_of_influence=49_000)
        res = modif([fake_bt, cth_clear], optional_datasets=[])
        np.testing.assert_allclose(res, fake_bt)
        with unittest.mock.patch("satpy.modifiers.parallax.resample_dataset") as smp:
            smp.side_effect = satpy.resample.base.resample_dataset
            modif([fake_bt, cth_clear], optional_datasets=[])
            assert smp.call_args_list[0].kwargs["radius_of_influence"] == 48_000
            assert smp.call_args_list[1].kwargs["radius_of_influence"] == 49_000

    def test_parallax_modifier_interface_with_cloud(self):
        """Test the modifier interface with a cloud.

        Test corresponds to a real bug encountered when using CTH data
        from NWCSAF-GEO, which created strange speckles in Africa (see
        https://github.com/pytroll/satpy/pull/1904#issuecomment-1011161623
        for an example).  Create fake CTH corresponding to NWCSAF-GEO area and
        BT corresponding to full disk SEVIRI, and test that no strange speckles
        occur.
        """
        from satpy.modifiers.parallax import ParallaxCorrectionModifier

        w_cth = 25
        h_cth = 15
        proj_dict = {"a": "6378137", "h": "35785863", "proj": "geos", "units": "m"}
        fake_area_cth = pyresample.create_area_def(
                area_id="test-area",
                projection=proj_dict,
                area_extent=(-2296808.75, 2785874.75, 2293808.25, 5570249.0),
                shape=(h_cth, w_cth))

        sz_bt = 20
        fake_area_bt = pyresample.create_area_def(
                "test-area-2",
                projection=proj_dict,
                area_extent=(-5567248.0742, -5513240.8172, 5513240.8172, 5567248.0742),
                shape=(sz_bt, sz_bt))

        (lons_cth, lats_cth) = fake_area_cth.get_lonlats()
        fake_cth_data = np.where(
                np.isfinite(lons_cth) & np.isfinite(lats_cth),
                15000,
                np.nan)

        (lons_bt, lats_bt) = fake_area_bt.get_lonlats()
        fake_bt_data = np.where(
                np.isfinite(lons_bt) & np.isfinite(lats_bt),
                np.linspace(200, 300, lons_bt.size).reshape(lons_bt.shape),
                np.nan)

        attrs = _get_attrs(0, 0)

        fake_bt = xr.DataArray(
                fake_bt_data,
                dims=("y", "x"),
                attrs={**attrs, "area": fake_area_bt})
        fake_cth = xr.DataArray(
                fake_cth_data,
                dims=("y", "x"),
                attrs={**attrs, "area": fake_area_cth})

        modif = ParallaxCorrectionModifier(
                name="parallax_corrected_dataset",
                prerequisites=[fake_bt, fake_cth],
                optional_prerequisites=[],
                search_radius=25_000)

        res = modif([fake_bt, fake_cth], optional_datasets=[])

        # with a constant cloud, a monotonically increasing BT should still
        # do so after parallax correction
        assert not (res.diff("x") < 0).any()

    @pytest.fixture
    def test_area(self, request):
        """Produce test area for parallax correction unit tests.

        Produce test area for the modifier-interface parallax correction unit
        tests.
        """
        extents = {
            "foroyar": [-861785.8867075047, 6820719.391005835, -686309.8124887547, 6954386.383193335],
            "ouagadougou": [-232482.90622750926, 1328206.360136668,
                            -114074.70310250926, 1422810.852324168],
            }
        where = request.param
        return pyresample.create_area_def(where, 4087, area_extent=extents[where], resolution=500)

    def _get_fake_cloud_datasets(self, test_area, cth, use_dask):
        """Return datasets for BT and CTH for fake cloud."""
        w_cloud = 20
        h_cloud = 5

        # location of cloud in uncorrected data
        lat_min_i = 155
        lat_max_i = lat_min_i + h_cloud
        lon_min_i = 140
        lon_max_i = lon_min_i + w_cloud

        fake_bt_data = np.linspace(
                270, 330, math.prod(test_area.shape), dtype="f8").reshape(
                        test_area.shape).round(2)
        fake_cth_data = np.full(test_area.shape, np.nan, dtype="f8")
        fake_bt_data[lat_min_i:lat_max_i, lon_min_i:lon_max_i] = np.linspace(
                180, 220, w_cloud*h_cloud).reshape(h_cloud, w_cloud).round(2)
        fake_cth_data[lat_min_i:lat_max_i, lon_min_i:lon_max_i] = cth

        if use_dask:
            fake_bt_data = da.array(fake_bt_data)
            fake_cth_data = da.array(fake_cth_data)

        attrs = _get_attrs(0, 0)

        fake_bt = xr.DataArray(
                fake_bt_data,
                dims=("y", "x"),
                attrs={**attrs, "area": test_area})

        fake_cth = xr.DataArray(
                fake_cth_data,
                dims=("y", "x"),
                attrs={**attrs, "area": test_area})

        cma = np.zeros(shape=fake_bt.shape, dtype="?")
        cma[lat_min_i:lat_max_i, lon_min_i:lon_max_i] = True

        return (fake_bt, fake_cth, cma)

    @pytest.mark.parametrize("test_area", ["foroyar", "ouagadougou"], indirect=["test_area"])
    def test_modifier_interface_fog_no_shift(self, test_area):
        """Test that fog isn't masked or shifted."""
        from satpy.modifiers.parallax import ParallaxCorrectionModifier

        (fake_bt, fake_cth, _) = self._get_fake_cloud_datasets(test_area, 50, use_dask=False)

        modif = ParallaxCorrectionModifier(
                name="parallax_corrected_dataset",
                prerequisites=[fake_bt, fake_cth],
                optional_prerequisites=[],
                debug_mode=True)

        res = modif([fake_bt, fake_cth], optional_datasets=[])

        assert np.isfinite(res).all()
        np.testing.assert_allclose(res, fake_bt)

    @pytest.mark.parametrize("cth", [7500, 15000])
    @pytest.mark.parametrize("use_dask", [True, False])
    @pytest.mark.parametrize("test_area", ["foroyar", "ouagadougou"], indirect=["test_area"])
    def test_modifier_interface_cloud_moves_to_observer(self, cth, use_dask, test_area):
        """Test that a cloud moves to the observer.

        With the modifier interface, use a high resolution area and test that
        pixels are moved in the direction of the observer and not away from it.
        """
        from satpy.modifiers.parallax import ParallaxCorrectionModifier

        (fake_bt, fake_cth, cma) = self._get_fake_cloud_datasets(test_area, cth, use_dask=use_dask)

        # location of cloud in corrected data
        # this may no longer be rectangular!
        dest_mask = np.zeros(shape=test_area.shape, dtype="?")
        cloud_location = {
                "foroyar": {
                    7500: (197, 202, 152, 172),
                    15000: (238, 243, 164, 184)},
                "ouagadougou": {
                    7500: (159, 164, 140, 160),
                    15000: (163, 168, 141, 161)}}
        (x_lo, x_hi, y_lo, y_hi) = cloud_location[test_area.area_id][cth]
        dest_mask[x_lo:x_hi, y_lo:y_hi] = True

        modif = ParallaxCorrectionModifier(
                name="parallax_corrected_dataset",
                prerequisites=[fake_bt, fake_cth],
                optional_prerequisites=[],
                debug_mode=True)

        res = modif([fake_bt, fake_cth], optional_datasets=[])

        assert fake_bt.attrs["area"] == test_area  # should not be changed
        assert res.attrs["area"] == fake_bt.attrs["area"]
        # confirm old cloud area now fill value
        # except where it overlaps with new cloud
        assert np.isnan(res.data[cma & (~dest_mask)]).all()
        # confirm rest of the area does not have fill values
        assert np.isfinite(res.data[~cma]).all()
        # confirm that rest of area pixel values did not change, except where
        # cloud arrived or originated
        delta = res - fake_bt
        assert (delta.data[~(cma | dest_mask)] == 0).all()
        # verify that cloud moved south.  Pointwise comparison might not work because
        # cloud may shrink.
        assert ((res.attrs["area"].get_lonlats()[1][dest_mask]).mean() <
                fake_bt.attrs["area"].get_lonlats()[1][cma].mean())
        # verify that most pixels at the new cloud location are indeed cloudy (calculate mean since some warm
        # surface pixels nearby a non-square cloud could be included in the square dest_mask
        assert res.data[dest_mask].mean() < 210


_test_yaml_code = """
sensor_name: visir

modifiers:
  parallax_corrected:
    modifier: !!python/name:satpy.modifiers.parallax.ParallaxCorrectionModifier
    prerequisites:
      - name: "ctth_alti"

composites:
  parallax_corrected_VIS006:
    compositor: !!python/name:satpy.composites.core.SingleBandCompositor
    prerequisites:
      - name: VIS006
        modifiers: [parallax_corrected]
"""


@pytest.mark.filterwarnings("ignore:Overlap checking not implemented:UserWarning")
class TestParallaxCorrectionSceneLoad:
    """Test that scene load interface works as expected."""

    @pytest.fixture
    def yaml_code(self):
        """Return YAML code for parallax_corrected_VIS006."""
        return _test_yaml_code

    @pytest.fixture
    def conf_file(self, yaml_code, tmp_path):
        """Produce a fake configuration file."""
        conf_file = tmp_path / "test.yaml"
        with conf_file.open(mode="wt", encoding="ascii") as fp:
            fp.write(yaml_code)
        return conf_file

    @pytest.fixture
    def fake_scene(self, yaml_code):
        """Produce fake scene and prepare fake composite config."""
        from satpy import Scene
        from satpy.dataset.dataid import WavelengthRange
        from satpy.tests.utils import make_dataid

        area = _get_fake_areas((0, 0), [5], 1)[0]
        sc = Scene()
        sc["VIS006"] = xr.DataArray(
            np.linspace(0, 99, 25).reshape(5, 5),
            dims=("y", "x"),
            attrs={
                "_satpy_id": make_dataid(
                    name="VIS006",
                    wavelength=WavelengthRange(min=0.56, central=0.635, max=0.71, unit="µm"),
                    resolution=3000,
                    calibration="reflectance",
                    modifiers=()),
                "modifiers": (),
                "sensor": "seviri",
                "area": area})
        sc["ctth_alti"] = xr.DataArray(
            np.linspace(0, 99, 25).reshape(5, 5),
            dims=("y", "x"),
            attrs={
                "_satpy_id": make_dataid(
                    name="ctth_alti",
                    resolution=3000,
                    modifiers=()),
                "modifiers": (),
                "sensor": {"seviri"},
                "platform_name": "Meteosat-11",
                "start_time": datetime.datetime(2022, 4, 12, 9, 0),
                "area": area})
        return sc

    def test_double_load(self, fake_scene, conf_file, fake_tle):
        """Test that loading corrected and uncorrected works correctly.

        When the modifier ``__call__`` method fails to call
        ``self.apply_modifier_info(new, old)`` and the original and
        parallax-corrected dataset are requested at the same time, the
        DataArrays differ but the underlying dask arrays have object identity,
        which in turn leads to both being parallax corrected.  This unit test
        confirms that there is no such object identity.
        """
        with unittest.mock.patch(
                "satpy.composites.config_loader.config_search_paths") as sccc, \
             unittest.mock.patch("pyorbital.tlefile.read") as plr:
            sccc.return_value = [os.fspath(conf_file)]
            plr.return_value = fake_tle
            fake_scene.load(["parallax_corrected_VIS006", "VIS006"])
            assert fake_scene["VIS006"] is not fake_scene["parallax_corrected_VIS006"]
        assert fake_scene["VIS006"].data is not fake_scene["parallax_corrected_VIS006"].data

    def test_no_compute(self, fake_scene, conf_file, fake_tle):
        """Test that no computation occurs."""
        from satpy.tests.utils import CustomScheduler
        with unittest.mock.patch(
                "satpy.composites.config_loader.config_search_paths") as sccc, \
             dask.config.set(scheduler=CustomScheduler(max_computes=0)), \
             unittest.mock.patch("pyorbital.tlefile.read") as plr:
            sccc.return_value = [os.fspath(conf_file)]
            plr.return_value = fake_tle
            fake_scene.load(["parallax_corrected_VIS006"])

    def test_enhanced_image(self, fake_scene, conf_file, fake_tle):
        """Test that image enhancement is the same."""
        with unittest.mock.patch(
                "satpy.composites.config_loader.config_search_paths") as sccc, \
             unittest.mock.patch("pyorbital.tlefile.read") as plr:
            sccc.return_value = [os.fspath(conf_file)]
            plr.return_value = fake_tle
            fake_scene.load(["parallax_corrected_VIS006", "VIS006"])
        im1 = get_enhanced_image(fake_scene["VIS006"])
        im2 = get_enhanced_image(fake_scene["parallax_corrected_VIS006"])
        assert im1.data.attrs["enhancement_history"] == im2.data.attrs["enhancement_history"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Nodes to build trees."""

from satpy.utils import get_logger

LOG = get_logger(__name__)
# Empty leaf used for marking composites with no prerequisites
EMPTY_LEAF_NAME = "__EMPTY_LEAF_SENTINEL__"


class MissingDependencies(RuntimeError):
    """Exception when dependencies are missing."""

    def __init__(self, missing_dependencies, *args, **kwargs):
        """Set up the exception."""
        super().__init__(*args, **kwargs)
        self.missing_dependencies = missing_dependencies

    def __str__(self):
        """Return the string representation of the exception."""
        prefix = super().__str__()
        unknown_str = ", ".join(map(str, self.missing_dependencies))
        return "{} {}".format(prefix, unknown_str)


class Node:
    """A node object."""

    def __init__(self, name, data=None):
        """Init the node object."""
        self.name = name
        self.data = data
        self.children = []
        self.parents = []

    def update_name(self, new_name):
        """Update 'name' property."""
        self.name = new_name

    @property
    def is_leaf(self):
        """Check if the node is a leaf."""
        return not self.children

    def flatten(self, d=None):
        """Flatten tree structure to a one level dictionary.

        Args:
            d (dict, Optional): output dictionary to update

        Returns:
            dict: Node.name -> Node. The returned dictionary includes the
                  current Node and all its children.

        """
        if d is None:
            d = {}
        if self.name is not None:
            d[self.name] = self
        for child in self.children:
            child.flatten(d=d)
        return d

    def copy(self, node_cache=None):
        """Make a copy of the node."""
        if node_cache and self.name in node_cache:
            return node_cache[self.name]

        if self.name is EMPTY_LEAF_NAME:
            return self

        s = self._copy_name_and_data(node_cache)
        for c in self.children:
            c = c.copy(node_cache=node_cache)
            s.add_child(c)
        if node_cache is not None:
            node_cache[s.name] = s
        return s

    def _copy_name_and_data(self, node_cache=None):
        return Node(self.name, self.data)

    def add_child(self, obj):
        """Add a child to the node."""
        self.children.append(obj)
        obj.parents.append(self)

    def __str__(self):
        """Display the node."""
        return self.display()

    def __repr__(self):
        """Generate a representation of the node."""
        return "<{} ({})>".format(self.__class__.__name__, repr(self.name))

    def __eq__(self, other):
        """Check equality."""
        return self.name == other.name

    def __hash__(self):
        """Generate the hash of the node."""
        return hash(self.name)

    def display(self, previous=0, include_data=False):
        """Display the node."""
        no_data = " (No Data)" if self.data is None else ""
        return (
            (" +" * previous) + str(self.name) + no_data + "\n" +
            "".join([child.display(previous + 1) for child in self.children]))

    def leaves(self, unique=True):
        """Get the leaves of the tree starting at this root."""
        if self.name is EMPTY_LEAF_NAME:
            return []
        elif not self.children:
            return [self]

        res = list()
        for child in self.children:
            for sub_child in child.leaves(unique=unique):
                if not unique or sub_child not in res:
                    res.append(sub_child)
        return res

    def trunk(self, unique=True, limit_children_to=None):
        """Get the trunk of the tree starting at this root."""
        # FIXME: uniqueness is not correct in `trunk` yet
        unique = False
        res = []
        if self.children and self.name is not EMPTY_LEAF_NAME:
            if self.name is not None:
                res.append(self)
            if limit_children_to is not None and self.name in limit_children_to:
                return res
            for child in self.children:
                for sub_child in child.trunk(unique=unique, limit_children_to=limit_children_to):
                    if not unique or sub_child not in res:
                        res.append(sub_child)
        return res


class CompositorNode(Node):
    """Implementation of a compositor-specific node."""

    def __init__(self, compositor):
        """Set up the node."""
        super().__init__(compositor.id, data=(compositor, [], []))

    def add_required_nodes(self, children):
        """Add nodes to the required field."""
        self.data[1].extend(children)

    @property
    def required_nodes(self):
        """Get the required nodes."""
        return self.data[1]

    def add_optional_nodes(self, children):
        """Add nodes to the optional field."""
        self.data[2].extend(children)

    @property
    def optional_nodes(self):
        """Get the optional nodes."""
        return self.data[2]

    @property
    def compositor(self):
        """Get the compositor."""
        return self.data[0]

    def _copy_name_and_data(self, node_cache=None):
        new_node = CompositorNode(self.compositor)
        new_required_nodes = [node.copy(node_cache) for node in self.required_nodes]
        new_node.add_required_nodes(new_required_nodes)
        new_optional_nodes = [node.copy(node_cache) for node in self.optional_nodes]
        new_node.add_optional_nodes(new_optional_nodes)
        # `comp.id` uses the compositor's attributes to compute itself
        # however, this node may have been updated by creation of the
        # composite. In order to not modify the compositor's attrs, we
        # overwrite the name here instead.
        new_node.name = self.name
        return new_node


class ReaderNode(Node):
    """Implementation of a storage-based node."""

    def __init__(self, unique_id, reader_name):
        """Set up the node."""
        super().__init__(unique_id, data={"reader_name": reader_name})

    def _copy_name_and_data(self, node_cache):
        return ReaderNode(self.name, self.data["reader_name"])

    @property
    def reader_name(self):
        """Get the name of the reader."""
        return self.data["reader_name"]



"""Functions and classes related to MultiScene functionality."""
from __future__ import annotations

from typing import Any

from satpy.utils import _import_and_warn_new_location

from ._multiscene import MultiScene  # noqa

IMPORT_PATHS = {
    "stack": "satpy.multiscene.blend_funcs",
    "temporal_rgb": "satpy.multiscene.blend_funcs",
    "timeseries": "satpy.multiscene.blend_funcs",
}


def __getattr__(name: str) -> Any:
    new_module = IMPORT_PATHS.get(name)

    if new_module is None:
        raise AttributeError(f"module {__name__} has no attribute '{name}'")

    return _import_and_warn_new_location(new_module, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Step for the real load-process-write tests."""

import fnmatch
import os
from tempfile import NamedTemporaryFile

import numpy as np
from behave import given, then, when
from PIL import Image


def fft_proj_rms(a1, a2):
    """Compute the RMS of differences between two images.

    Compute the RMS of differences between two FFT vectors of a1
    and projection of FFT vectors of a2.
    This metric is sensitive to large scale changes and image noise but
    insensitive to small rendering differences.
    """
    ms = 0

    # for i in range(a1.shape[-1]):
    fr1 = np.fft.rfftn(a1)
    fr2 = np.fft.rfftn(a2)

    ps1 = np.log10(fr1 * fr1.conj()).real
    ps2 = np.log10(fr2 * fr2.conj()).real

    p1 = np.arctan2(fr1.imag, fr1.real)
    p2 = np.arctan2(fr2.imag, fr2.real)

    theta = p2 - p1
    l_factor = ps2 * np.cos(theta)
    ms += np.sum(((l_factor - ps1) ** 2)) / float(ps1.size)

    rms = np.sqrt(ms)

    return rms


def assert_images_match(image1, image2, threshold=0.1):
    """Assert that images are matching."""
    img1 = np.asarray(Image.open(image1))
    img2 = np.asarray(Image.open(image2))
    rms = fft_proj_rms(img1, img2)
    assert rms <= threshold, "Images {0} and {1} don't match: {2}".format(
        image1, image2, rms)


def get_all_files(directory, pattern):
    """Find all files matching *pattern* under ``directory``."""
    matches = []
    for root, _, filenames in os.walk(directory):
        for filename in fnmatch.filter(filenames, pattern):
            matches.append(os.path.join(root, filename))
    return matches


def before_all(context):
    """Enable satpy debugging."""
    if not context.config.log_capture:
        from satpy.utils import debug_on
        debug_on()


@given(u"{dformat} data is available")
def step_impl_input_files_exists(context, dformat):
    """Check that input data exists on disk."""
    data_path = os.path.join("test_data", dformat)
    data_available = os.path.exists(data_path)
    if not data_available:
        context.scenario.skip(reason="No test data available for " + dformat)
    else:
        context.dformat = dformat
        context.data_path = data_path


@when(u"the user loads the {composite} composite")
def step_impl_create_scene_and_load_single(context, composite):
    """Create a Scene and load a single composite."""
    from satpy import Scene
    scn = Scene(reader=context.dformat,
                filenames=get_all_files(os.path.join(context.data_path, "data"),
                                        "*"))
    scn.load([composite])
    context.scn = scn
    context.composite = composite


@when(u"the user resamples the data to {area}")
def step_impl_resample_scene(context, area):
    """Resample the scene to an area or use the native resampler."""
    if area != "-":
        context.lscn = context.scn.resample(area)
    else:
        context.lscn = context.scn.resample(resampler="native")
    context.area = area


@when(u"the user saves the composite to disk")
def step_impl_save_to_png(context):
    """Call Scene.save_dataset to write a PNG image."""
    with NamedTemporaryFile(suffix=".png", delete=False) as tmp_file:
        context.lscn.save_dataset(context.composite, filename=tmp_file.name)
        context.new_filename = tmp_file.name


@then(u"the resulting image should match the reference image")
def step_impl_compare_two_png_images(context):
    """Compare two PNG image files."""
    if context.area == "-":
        ref_filename = context.composite + ".png"
    else:
        ref_filename = context.composite + "_" + context.area + ".png"
    ref_filename = os.path.join(context.data_path, "ref", ref_filename)
    assert os.path.exists(ref_filename), "Missing reference file."
    assert_images_match(ref_filename, context.new_filename)
    os.remove(context.new_filename)



# Copyright (c) 2015-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Test objects and functions in the dataset module."""

import datetime as dt
import unittest

import numpy as np
import pytest

from satpy.dataset.dataid import DataID, DataQuery, ModifierTuple, WavelengthRange, minimal_default_keys_config
from satpy.readers.pmw_channels_definitions import FrequencyDoubleSideBand, FrequencyQuadrupleSideBand, FrequencyRange
from satpy.tests.utils import make_cid, make_dataid, make_dsq


class TestDataID(unittest.TestCase):
    """Test DataID object creation and other methods."""

    def test_basic_init(self):
        """Test basic ways of creating a DataID."""
        from satpy.dataset.dataid import DataID
        from satpy.dataset.dataid import default_id_keys_config as dikc
        from satpy.dataset.dataid import minimal_default_keys_config as mdkc

        did = DataID(dikc, name="a")
        assert did["name"] == "a"
        assert did["modifiers"] == tuple()
        DataID(dikc, name="a", wavelength=0.86)
        DataID(dikc, name="a", resolution=1000)
        DataID(dikc, name="a", calibration="radiance")
        DataID(dikc, name="a", wavelength=0.86, resolution=250,
               calibration="radiance")
        DataID(dikc, name="a", wavelength=0.86, resolution=250,
               calibration="radiance", modifiers=("sunz_corrected",))
        with pytest.raises(ValueError, match="Required field name missing."):
            DataID(dikc, wavelength=0.86)
        did = DataID(mdkc, name="comp24", resolution=500)
        assert did["resolution"] == 500

    def test_init_bad_modifiers(self):
        """Test that modifiers are a tuple."""
        from satpy.dataset.dataid import DataID
        from satpy.dataset.dataid import default_id_keys_config as dikc
        with pytest.raises(TypeError):
            DataID(dikc, name="a", modifiers="str")

    def test_compare_no_wl(self):
        """Compare fully qualified wavelength ID to no wavelength ID."""
        from satpy.dataset.dataid import DataID
        from satpy.dataset.dataid import default_id_keys_config as dikc
        d1 = DataID(dikc, name="a", wavelength=(0.1, 0.2, 0.3))
        d2 = DataID(dikc, name="a", wavelength=None)

        # this happens when sorting IDs during dependency checks
        assert not (d1 < d2)
        assert d2 < d1

    def test_bad_calibration(self):
        """Test that asking for a bad calibration fails."""
        from satpy.dataset.dataid import DataID
        from satpy.dataset.dataid import default_id_keys_config as dikc
        with pytest.raises(ValueError, match="_bad_ invalid value for <enum 'calibration'>"):
            DataID(dikc, name="C05", calibration="_bad_")

    def test_is_modified(self):
        """Test that modifications are detected properly."""
        from satpy.dataset.dataid import DataID
        from satpy.dataset.dataid import default_id_keys_config as dikc
        d1 = DataID(dikc, name="a", wavelength=(0.1, 0.2, 0.3), modifiers=("hej",))
        d2 = DataID(dikc, name="a", wavelength=(0.1, 0.2, 0.3), modifiers=tuple())

        assert d1.is_modified()
        assert not d2.is_modified()

    def test_create_less_modified_query(self):
        """Test that modifications are popped correctly."""
        from satpy.dataset.dataid import DataID
        from satpy.dataset.dataid import default_id_keys_config as dikc
        d1 = DataID(dikc, name="a", wavelength=(0.1, 0.2, 0.3), modifiers=("hej",))
        d2 = DataID(dikc, name="a", wavelength=(0.1, 0.2, 0.3), modifiers=tuple())

        assert not d1.create_less_modified_query()["modifiers"]
        assert not d2.create_less_modified_query()["modifiers"]


class TestCombineMetadata(unittest.TestCase):
    """Test how metadata is combined."""

    def setUp(self):
        """Set up the test case."""
        # The times need to be in ascending order (oldest first)
        self.start_time_dts = (
            {"start_time": dt.datetime(2018, 2, 1, 11, 58, 0)},
            {"start_time": dt.datetime(2018, 2, 1, 11, 59, 0)},
            {"start_time": dt.datetime(2018, 2, 1, 12, 0, 0)},
            {"start_time": dt.datetime(2018, 2, 1, 12, 1, 0)},
            {"start_time": dt.datetime(2018, 2, 1, 12, 2, 0)},
        )
        self.end_time_dts = (
            {"end_time": dt.datetime(2018, 2, 1, 11, 58, 0)},
            {"end_time": dt.datetime(2018, 2, 1, 11, 59, 0)},
            {"end_time": dt.datetime(2018, 2, 1, 12, 0, 0)},
            {"end_time": dt.datetime(2018, 2, 1, 12, 1, 0)},
            {"end_time": dt.datetime(2018, 2, 1, 12, 2, 0)},
        )
        self.other_time_dts = (
            {"other_time": dt.datetime(2018, 2, 1, 11, 58, 0)},
            {"other_time": dt.datetime(2018, 2, 1, 11, 59, 0)},
            {"other_time": dt.datetime(2018, 2, 1, 12, 0, 0)},
            {"other_time": dt.datetime(2018, 2, 1, 12, 1, 0)},
            {"other_time": dt.datetime(2018, 2, 1, 12, 2, 0)},
        )
        self.start_time_dts_with_none = (
            {"start_time": None},
            {"start_time": dt.datetime(2018, 2, 1, 11, 59, 0)},
            {"start_time": dt.datetime(2018, 2, 1, 12, 0, 0)},
            {"start_time": dt.datetime(2018, 2, 1, 12, 1, 0)},
            {"start_time": dt.datetime(2018, 2, 1, 12, 2, 0)},
        )
        self.end_time_dts_with_none = (
            {"end_time": dt.datetime(2018, 2, 1, 11, 58, 0)},
            {"end_time": dt.datetime(2018, 2, 1, 11, 59, 0)},
            {"end_time": dt.datetime(2018, 2, 1, 12, 0, 0)},
            {"end_time": dt.datetime(2018, 2, 1, 12, 1, 0)},
            {"end_time": None},
        )

    def test_average_datetimes(self):
        """Test the average_datetimes helper function."""
        from satpy.dataset.metadata import average_datetimes
        dts = (
            dt.datetime(2018, 2, 1, 11, 58, 0),
            dt.datetime(2018, 2, 1, 11, 59, 0),
            dt.datetime(2018, 2, 1, 12, 0, 0),
            dt.datetime(2018, 2, 1, 12, 1, 0),
            dt.datetime(2018, 2, 1, 12, 2, 0),
        )
        ret = average_datetimes(dts)
        assert dts[2] == ret

    def test_combine_start_times(self):
        """Test the combine_metadata with start times."""
        from satpy.dataset.metadata import combine_metadata
        ret = combine_metadata(*self.start_time_dts)
        assert ret["start_time"] == self.start_time_dts[0]["start_time"]

    def test_combine_end_times(self):
        """Test the combine_metadata with end times."""
        from satpy.dataset.metadata import combine_metadata
        ret = combine_metadata(*self.end_time_dts)
        assert ret["end_time"] == self.end_time_dts[-1]["end_time"]

    def test_combine_start_times_with_none(self):
        """Test the combine_metadata with start times when there's a None included."""
        from satpy.dataset.metadata import combine_metadata
        ret = combine_metadata(*self.start_time_dts_with_none)
        assert ret["start_time"] == self.start_time_dts_with_none[1]["start_time"]

    def test_combine_end_times_with_none(self):
        """Test the combine_metadata with end times when there's a None included."""
        from satpy.dataset.metadata import combine_metadata
        ret = combine_metadata(*self.end_time_dts_with_none)
        assert ret["end_time"] == self.end_time_dts_with_none[-2]["end_time"]

    def test_combine_other_times(self):
        """Test the combine_metadata with other time values than start or end times."""
        from satpy.dataset.metadata import combine_metadata
        ret = combine_metadata(*self.other_time_dts)
        assert ret["other_time"] == self.other_time_dts[2]["other_time"]

    def test_combine_arrays(self):
        """Test the combine_metadata with arrays."""
        from numpy import arange, ones
        from xarray import DataArray

        from satpy.dataset.metadata import combine_metadata
        dts = [
            {"quality": (arange(25) % 2).reshape(5, 5).astype("?")},
            {"quality": (arange(1, 26) % 3).reshape(5, 5).astype("?")},
            {"quality": ones((5, 5,), "?")},
        ]
        assert "quality" not in combine_metadata(*dts)
        dts2 = [{"quality": DataArray(d["quality"])} for d in dts]
        assert "quality" not in combine_metadata(*dts2)
        # the ancillary_variables attribute is actually a list of data arrays
        dts3 = [{"quality": [d["quality"]]} for d in dts]
        assert "quality" not in combine_metadata(*dts3)
        # check cases with repeated arrays
        dts4 = [
            {"quality": dts[0]["quality"]},
            {"quality": dts[0]["quality"]},
        ]
        assert "quality" in combine_metadata(*dts4)
        dts5 = [
            {"quality": dts3[0]["quality"]},
            {"quality": dts3[0]["quality"]},
        ]
        assert "quality" in combine_metadata(*dts5)
        # check with other types
        dts6 = [
            DataArray(arange(5), attrs=dts[0]),
            DataArray(arange(5), attrs=dts[0]),
            DataArray(arange(5), attrs=dts[1]),
            object()
        ]
        assert "quality" not in combine_metadata(*dts6)

    def test_combine_lists_identical(self):
        """Test combine metadata with identical lists."""
        from satpy.dataset.metadata import combine_metadata
        metadatas = [
            {"prerequisites": [1, 2, 3, 4]},
            {"prerequisites": [1, 2, 3, 4]},
        ]
        res = combine_metadata(*metadatas)
        assert res["prerequisites"] == [1, 2, 3, 4]

    def test_combine_lists_same_size_diff_values(self):
        """Test combine metadata with lists with different values."""
        from satpy.dataset.metadata import combine_metadata
        metadatas = [
            {"prerequisites": [1, 2, 3, 4]},
            {"prerequisites": [1, 2, 3, 5]},
        ]
        res = combine_metadata(*metadatas)
        assert "prerequisites" not in res

    def test_combine_lists_different_size(self):
        """Test combine metadata with different size lists."""
        from satpy.dataset.metadata import combine_metadata
        metadatas = [
            {"prerequisites": [1, 2, 3, 4]},
            {"prerequisites": []},
        ]
        res = combine_metadata(*metadatas)
        assert "prerequisites" not in res

        metadatas = [
            {"prerequisites": [1, 2, 3, 4]},
            {"prerequisites": [1, 2, 3]},
        ]
        res = combine_metadata(*metadatas)
        assert "prerequisites" not in res

    def test_combine_identical_numpy_scalars(self):
        """Test combining identical fill values."""
        from satpy.dataset.metadata import combine_metadata
        test_metadata = [{"_FillValue": np.uint16(42)}, {"_FillValue": np.uint16(42)}]
        assert combine_metadata(*test_metadata) == {"_FillValue": 42}

    def test_combine_empty_metadata(self):
        """Test combining empty metadata."""
        from satpy.dataset.metadata import combine_metadata
        test_metadata = [{}, {}]
        assert combine_metadata(*test_metadata) == {}

    def test_combine_nans(self):
        """Test combining nan fill values."""
        from satpy.dataset.metadata import combine_metadata
        test_metadata = [{"_FillValue": np.nan}, {"_FillValue": np.nan}]
        assert combine_metadata(*test_metadata) == {"_FillValue": np.nan}

    def test_combine_numpy_arrays(self):
        """Test combining values that are numpy arrays."""
        from satpy.dataset.metadata import combine_metadata
        test_metadata = [{"valid_range": np.array([0., 0.00032], dtype=np.float32)},
                         {"valid_range": np.array([0., 0.00032], dtype=np.float32)},
                         {"valid_range": np.array([0., 0.00032], dtype=np.float32)}]
        result = combine_metadata(*test_metadata)
        assert np.allclose(result["valid_range"], np.array([0., 0.00032], dtype=np.float32))

    def test_combine_dask_arrays(self):
        """Test combining values that are dask arrays."""
        import dask.array as da

        from satpy.dataset.metadata import combine_metadata
        test_metadata = [{"valid_range": da.from_array(np.array([0., 0.00032], dtype=np.float32))},
                         {"valid_range": da.from_array(np.array([0., 0.00032], dtype=np.float32))}]
        result = combine_metadata(*test_metadata)
        assert "valid_range" not in result

    def test_combine_xarray_arrays(self):
        """Test combining values that are non daskified xarray dataArrays."""
        import xarray as xr

        from satpy.dataset.metadata import combine_metadata

        test_metadata = [{"valid_range": xr.DataArray(np.array([0., 0.00032], dtype=np.float32),
                         attrs={"_FillValue": -9999})},
                         {"valid_range": xr.DataArray(np.array([0., 0.00032], dtype=np.float32),
                         attrs={"_FillValue": -9999})},
                         {"valid_range": xr.DataArray(np.array([0., 0.00032], dtype=np.float32),
                         attrs={"_FillValue": -9999})}]
        result = combine_metadata(*test_metadata)
        assert np.allclose(result["valid_range"], xr.DataArray(np.array([0., 0.00032], dtype=np.float32),
                         attrs={"_FillValue": -9999}))

    def test_combine_real_world_mda(self):
        """Test with real data."""
        mda_objects = ({"_FillValue": np.nan,
                        "valid_range": np.array([0., 0.00032], dtype=np.float32),
                        "ancillary_variables": ["cpp_status_flag",
                                                "cpp_conditions",
                                                "cpp_quality",
                                                "cpp_reff_pal",
                                                "-"],
                        "platform_name": "NOAA-20",
                        "sensor": {"viirs"},
                        "raw_metadata": {"foo": {"bar": np.array([1, 2, 3])}}},
                       {"_FillValue": np.nan,
                        "valid_range": np.array([0., 0.00032], dtype=np.float32),
                        "ancillary_variables": ["cpp_status_flag",
                                                "cpp_conditions",
                                                "cpp_quality",
                                                "cpp_reff_pal",
                                                "-"],
                        "platform_name": "NOAA-20",
                        "sensor": {"viirs"},
                        "raw_metadata": {"foo": {"bar": np.array([1, 2, 3])}}})

        expected = {"_FillValue": np.nan,
                    "valid_range": np.array([0., 0.00032], dtype=np.float32),
                    "ancillary_variables": ["cpp_status_flag",
                                            "cpp_conditions",
                                            "cpp_quality",
                                            "cpp_reff_pal",
                                            "-"],
                    "platform_name": "NOAA-20",
                    "sensor": {"viirs"},
                    "raw_metadata": {"foo": {"bar": np.array([1, 2, 3])}}}

        from satpy.dataset.metadata import combine_metadata
        result = combine_metadata(*mda_objects)
        assert np.allclose(result.pop("_FillValue"), expected.pop("_FillValue"), equal_nan=True)
        assert np.allclose(result.pop("valid_range"), expected.pop("valid_range"))
        np.testing.assert_equal(result.pop("raw_metadata"),
                                expected.pop("raw_metadata"))
        assert result == expected

    def test_combine_one_metadata_object(self):
        """Test combining one metadata object."""
        mda_objects = ({"_FillValue": np.nan,
                        "valid_range": np.array([0., 0.00032], dtype=np.float32),
                        "ancillary_variables": ["cpp_status_flag",
                                                "cpp_conditions",
                                                "cpp_quality",
                                                "cpp_reff_pal",
                                                "-"],
                        "platform_name": "NOAA-20",
                        "sensor": {"viirs"}},)

        expected = {"_FillValue": np.nan,
                    "valid_range": np.array([0., 0.00032], dtype=np.float32),
                    "ancillary_variables": ["cpp_status_flag",
                                            "cpp_conditions",
                                            "cpp_quality",
                                            "cpp_reff_pal",
                                            "-"],
                    "platform_name": "NOAA-20",
                    "sensor": {"viirs"}}

        from satpy.dataset.metadata import combine_metadata
        result = combine_metadata(*mda_objects)
        assert np.allclose(result.pop("_FillValue"), expected.pop("_FillValue"), equal_nan=True)
        assert np.allclose(result.pop("valid_range"), expected.pop("valid_range"))
        assert result == expected


def test_combine_dicts_close():
    """Test combination of dictionaries whose values are close."""
    from satpy.dataset.metadata import combine_metadata
    attrs = {
        "raw_metadata": {
            "a": 1,
            "b": "foo",
            "c": [1, 2, 3],
            "d": {
                "e": np.str_("bar"),
                "f": dt.datetime(2020, 1, 1, 12, 15, 30),
                "g": np.array([1, 2, 3]),
            },
            "h": np.array([dt.datetime(2020, 1, 1), dt.datetime(2020, 1, 1)])
        }
    }
    attrs_close = {
        "raw_metadata": {
            "a": 1 + 1E-12,
            "b": "foo",
            "c": np.array([1, 2, 3]) + 1E-12,
            "d": {
                "e": np.str_("bar"),
                "f": dt.datetime(2020, 1, 1, 12, 15, 30),
                "g": np.array([1, 2, 3]) + 1E-12
            },
            "h": np.array([dt.datetime(2020, 1, 1), dt.datetime(2020, 1, 1)])
        }
    }
    test_metadata = [attrs, attrs_close]
    result = combine_metadata(*test_metadata)
    assert result == attrs


@pytest.mark.parametrize(
    "test_mda",
    [
        # a/b/c/d different
        {"a": np.array([1, 2, 3]), "d": 123},
        {"a": {"b": np.array([4, 5, 6]), "c": 1.0}, "d": "foo"},
        {"a": {"b": np.array([1, 2, 3]), "c": 2.0}, "d": "foo"},
        {"a": {"b": np.array([1, 2, 3]), "c": 1.0}, "d": "bar"},
        # a/b/c/d type different
        np.array([1, 2, 3]),
        {"a": {"b": "baz", "c": 1.0}, "d": "foo"},
        {"a": {"b": np.array([1, 2, 3]), "c": "baz"}, "d": "foo"},
        {"a": {"b": np.array([1, 2, 3]), "c": 1.0}, "d": 1.0}
    ]
)
def test_combine_dicts_different(test_mda):
    """Test combination of dictionaries differing in various ways."""
    from satpy.dataset.metadata import combine_metadata
    mda = {"a": {"b": np.array([1, 2, 3]), "c": 1.0}, "d": "foo"}
    test_metadata = [{"raw_metadata": mda}, {"raw_metadata": test_mda}]
    result = combine_metadata(*test_metadata)
    assert not result


def test_dataid():
    """Test the DataID object."""
    from satpy.dataset.dataid import DataID, ModifierTuple, ValueList, WavelengthRange

    # Check that enum is translated to type.
    did = make_dataid()
    assert issubclass(did._id_keys["calibration"]["type"], ValueList)
    assert "enum" not in did._id_keys["calibration"]

    # Check that None is never a valid value
    did = make_dataid(name="cheese_shops", resolution=None)
    assert "resolution" not in did
    assert "None" not in did.__repr__()
    with pytest.raises(ValueError, match="Required field name missing."):
        make_dataid(name=None, resolution=1000)

    # Check that defaults are applied correctly
    assert did["modifiers"] == ModifierTuple()

    # Check that from_dict creates a distinct instance...
    did2 = did.from_dict(dict(name="cheese_shops", resolution=None))
    assert did is not did2
    # ...But is equal
    assert did2 == did

    # Check that the instance is immutable
    with pytest.raises(TypeError):
        did["resolution"] = 1000

    # Check that a missing required field crashes
    with pytest.raises(ValueError, match="Required field name missing."):
        make_dataid(resolution=1000)

    # Check to_dict
    assert did.to_dict() == dict(name="cheese_shops", modifiers=tuple())

    # Check repr
    did = make_dataid(name="VIS008", resolution=111)
    assert repr(did) == "DataID(name='VIS008', resolution=111, modifiers=())"

    # Check inequality
    default_id_keys_config = {"name": None,
                              "wavelength": {
                                  "type": WavelengthRange,
                              },
                              "resolution": None,
                              "calibration": {
                                  "enum": [
                                      "reflectance",
                                      "brightness_temperature",
                                      "radiance",
                                      "counts"
                                  ]
                              },
                              "modifiers": {
                                  "default": ModifierTuple(),
                                  "type": ModifierTuple,
                              },
                              }
    assert DataID(default_id_keys_config, wavelength=10) != DataID(default_id_keys_config, name="VIS006")


def test_dataid_equal_if_enums_different():
    """Check that dataids with different enums but same items are equal."""
    from satpy.dataset.dataid import DataID, ModifierTuple, WavelengthRange
    id_keys_config1 = {"name": None,
                       "wavelength": {
                           "type": WavelengthRange,
                       },
                       "resolution": None,
                       "calibration": {
                           "enum": [
                               "c1",
                               "c2",
                               "c3",
                           ]
                       },
                       "modifiers": {
                           "default": ModifierTuple(),
                           "type": ModifierTuple,
                       },
                       }

    id_keys_config2 = {"name": None,
                       "wavelength": {
                           "type": WavelengthRange,
                       },
                       "resolution": None,
                       "calibration": {
                           "enum": [
                               "c1",
                               "c1.5",
                               "c2",
                               "c2.5",
                               "c3"
                           ]
                       },
                       "modifiers": {
                           "default": ModifierTuple(),
                           "type": ModifierTuple,
                       },
                       }
    assert DataID(id_keys_config1, name="ni", calibration="c2") == DataID(id_keys_config2, name="ni", calibration="c2")


def test_dataid_copy():
    """Test copying a DataID."""
    from copy import deepcopy

    from satpy.dataset.dataid import DataID
    from satpy.dataset.dataid import default_id_keys_config as dikc

    did = DataID(dikc, name="a", resolution=1000)
    did2 = deepcopy(did)
    assert did2 == did
    assert did2.id_keys == did.id_keys


def test_dataid_pickle():
    """Test dataid pickling roundtrip."""
    import pickle

    from satpy.tests.utils import make_dataid
    did = make_dataid(name="hi", wavelength=(10, 11, 12), resolution=1000, calibration="radiance")
    assert did == pickle.loads(pickle.dumps(did))


def test_dataid_elements_picklable():
    """Test individual elements of DataID can be pickled.

    In some cases, like in the base reader classes, the elements of a DataID
    are extracted and stored in a separate dictionary. This means that the
    internal/fancy pickle handling of DataID does not play a part.

    """
    import pickle

    from satpy.tests.utils import make_dataid
    did = make_dataid(name="hi", wavelength=(10, 11, 12), resolution=1000, calibration="radiance")
    for value in did.values():
        pickled_value = pickle.loads(pickle.dumps(value))
        assert value == pickled_value


class TestDataQuery:
    """Test case for data queries."""

    def test_dataquery(self):
        """Test DataQuery objects."""
        from satpy.dataset import DataQuery

        DataQuery(name="cheese_shops")

        # Check repr
        did = DataQuery(name="VIS008", resolution=111)
        assert repr(did) == "DataQuery(name='VIS008', resolution=111)"

        # Check inequality
        assert DataQuery(wavelength=10) != DataQuery(name="VIS006")

    def test_is_modified(self):
        """Test that modifications are detected properly."""
        from satpy.dataset import DataQuery
        d1 = DataQuery(name="a", wavelength=0.2, modifiers=("hej",))
        d2 = DataQuery(name="a", wavelength=0.2, modifiers=tuple())

        assert d1.is_modified()
        assert not d2.is_modified()

    def test_create_less_modified_query(self):
        """Test that modifications are popped correctly."""
        from satpy.dataset import DataQuery
        d1 = DataQuery(name="a", wavelength=0.2, modifiers=("hej",))
        d2 = DataQuery(name="a", wavelength=0.2, modifiers=tuple())

        assert not d1.create_less_modified_query()["modifiers"]
        assert not d2.create_less_modified_query()["modifiers"]


class TestIDQueryInteractions(unittest.TestCase):
    """Test the interactions between DataIDs and DataQuerys."""

    def setUp(self) -> None:
        """Set up the test case."""
        self.default_id_keys_config = {
            "name": {
                "required": True,
            },
            "wavelength": {
                "type": WavelengthRange,
            },
            "resolution": None,
            "calibration": {
                "enum": [
                    "reflectance",
                    "brightness_temperature",
                    "radiance",
                    "counts"
                ]
            },
            "modifiers": {
                "default": ModifierTuple(),
                "type": ModifierTuple,
            },
        }

    def test_hash_equality(self):
        """Test hash equality."""
        dq = DataQuery(modifiers=tuple(), name="cheese_shops")
        did = DataID(self.default_id_keys_config, name="cheese_shops")
        assert hash(dq) == hash(did)

    def test_id_filtering(self):
        """Check did filtering."""
        dq = DataQuery(modifiers=tuple(), name="cheese_shops")
        did = DataID(self.default_id_keys_config, name="cheese_shops")
        did2 = DataID(self.default_id_keys_config, name="ni")
        res = dq.filter_dataids([did2, did])
        assert len(res) == 1
        assert res[0] == did

        dataid_container = [DataID(self.default_id_keys_config,
                                   name="ds1",
                                   resolution=250,
                                   calibration="reflectance",
                                   modifiers=tuple())]
        dq = DataQuery(wavelength=0.22, modifiers=tuple())
        assert len(dq.filter_dataids(dataid_container)) == 0
        dataid_container = [DataID(minimal_default_keys_config,
                                   name="natural_color")]
        dq = DataQuery(name="natural_color", resolution=250)
        assert len(dq.filter_dataids(dataid_container)) == 1

        dq = make_dsq(wavelength=0.22, modifiers=("mod1",))
        did = make_cid(name="static_image")
        assert len(dq.filter_dataids([did])) == 0

    def test_inequality(self):
        """Check (in)equality."""
        assert DataQuery(wavelength=10) != DataID(self.default_id_keys_config, name="VIS006")

    def test_sort_dataids(self):
        """Check dataid sorting."""
        dq = DataQuery(name="cheese_shops", wavelength=2, modifiers="*")
        did = DataID(self.default_id_keys_config, name="cheese_shops", wavelength=(1, 2, 3))
        did2 = DataID(self.default_id_keys_config, name="cheese_shops", wavelength=(1.1, 2.1, 3.1))
        dsids, distances = dq.sort_dataids([did2, did])
        assert list(dsids) == [did, did2]
        assert np.allclose(distances, [0, 0.1])

        dq = DataQuery(name="cheese_shops")
        did = DataID(self.default_id_keys_config, name="cheese_shops", resolution=200)
        did2 = DataID(self.default_id_keys_config, name="cheese_shops", resolution=400)
        dsids, distances = dq.sort_dataids([did2, did])
        assert list(dsids) == [did, did2]
        assert distances[0] < distances[1]

        did = DataID(self.default_id_keys_config, name="cheese_shops", calibration="counts")
        did2 = DataID(self.default_id_keys_config, name="cheese_shops", calibration="reflectance")
        dsids, distances = dq.sort_dataids([did2, did])
        assert list(dsids) == [did2, did]
        assert distances[0] < distances[1]

        did = DataID(self.default_id_keys_config, name="cheese_shops", modifiers=tuple())
        did2 = DataID(self.default_id_keys_config, name="cheese_shops", modifiers=tuple(["out_of_stock"]))
        dsids, distances = dq.sort_dataids([did2, did])
        assert list(dsids) == [did, did2]
        assert distances[0] < distances[1]

    def test_sort_dataids_with_different_set_of_keys(self):
        """Check sorting data ids when the query has a different set of keys."""
        dq = DataQuery(name="solar_zenith_angle", calibration="reflectance")
        dids = [DataID(self.default_id_keys_config, name="solar_zenith_angle", resolution=1000, modifiers=()),
                DataID(self.default_id_keys_config, name="solar_zenith_angle", resolution=500, modifiers=()),
                DataID(self.default_id_keys_config, name="solar_zenith_angle", resolution=250, modifiers=())]
        dsids, distances = dq.sort_dataids(dids)
        assert distances[0] < distances[1]
        assert distances[1] < distances[2]

    def test_seviri_hrv_has_priority_over_vis008(self):
        """Check that the HRV channel has priority over VIS008 when querying 0.8µm."""
        dids = [DataID(self.default_id_keys_config, name="HRV",
                       wavelength=WavelengthRange(min=0.5, central=0.7, max=0.9, unit="µm"), resolution=1000.134348869,
                       calibration="reflectance", modifiers=()),
                DataID(self.default_id_keys_config, name="HRV",
                       wavelength=WavelengthRange(min=0.5, central=0.7, max=0.9, unit="µm"), resolution=1000.134348869,
                       calibration="radiance", modifiers=()),
                DataID(self.default_id_keys_config, name="HRV",
                       wavelength=WavelengthRange(min=0.5, central=0.7, max=0.9, unit="µm"), resolution=1000.134348869,
                       calibration="counts", modifiers=()),
                DataID(self.default_id_keys_config, name="VIS006",
                       wavelength=WavelengthRange(min=0.56, central=0.635, max=0.71, unit="µm"),
                       resolution=3000.403165817, calibration="reflectance", modifiers=()),
                DataID(self.default_id_keys_config, name="VIS006",
                       wavelength=WavelengthRange(min=0.56, central=0.635, max=0.71, unit="µm"),
                       resolution=3000.403165817, calibration="radiance", modifiers=()),
                DataID(self.default_id_keys_config, name="VIS006",
                       wavelength=WavelengthRange(min=0.56, central=0.635, max=0.71, unit="µm"),
                       resolution=3000.403165817, calibration="counts", modifiers=()),
                DataID(self.default_id_keys_config, name="VIS008",
                       wavelength=WavelengthRange(min=0.74, central=0.81, max=0.88, unit="µm"),
                       resolution=3000.403165817, calibration="reflectance", modifiers=()),
                DataID(self.default_id_keys_config, name="VIS008",
                       wavelength=WavelengthRange(min=0.74, central=0.81, max=0.88, unit="µm"),
                       resolution=3000.403165817, calibration="radiance", modifiers=()),
                DataID(self.default_id_keys_config, name="VIS008",
                       wavelength=WavelengthRange(min=0.74, central=0.81, max=0.88, unit="µm"),
                       resolution=3000.403165817, calibration="counts", modifiers=())]
        dq = DataQuery(wavelength=0.8)
        res, distances = dq.sort_dataids(dids)
        assert res[0]["name"] == "HRV"


def test_frequency_quadruple_side_band_class_method_convert():
    """Test the frequency double side band object: test the class method convert."""
    frq_qdsb = FrequencyQuadrupleSideBand(57, 0.322, 0.05, 0.036)

    res = frq_qdsb.convert(57.37)
    assert res == 57.37

    res = frq_qdsb.convert({"central": 57.0, "side": 0.322, "sideside": 0.05, "bandwidth": 0.036})
    assert res == FrequencyQuadrupleSideBand(57, 0.322, 0.05, 0.036)


def test_frequency_quadruple_side_band_channel_str():
    """Test the frequency quadruple side band object: test the band description."""
    frq_qdsb1 = FrequencyQuadrupleSideBand(57.0, 0.322, 0.05, 0.036)
    frq_qdsb2 = FrequencyQuadrupleSideBand(57000, 322, 50, 36, "MHz")

    assert str(frq_qdsb1) == "central=57.0 GHz ±0.322 ±0.05 width=0.036 GHz"
    assert str(frq_qdsb2) == "central=57000 MHz ±322 ±50 width=36 MHz"


def test_frequency_quadruple_side_band_channel_equality():
    """Test the frequency quadruple side band object: check if two bands are 'equal'."""
    frq_qdsb = FrequencyQuadrupleSideBand(57, 0.322, 0.05, 0.036)
    assert frq_qdsb is not None
    assert frq_qdsb < FrequencyQuadrupleSideBand(57, 0.322, 0.05, 0.04)
    assert frq_qdsb < FrequencyQuadrupleSideBand(58, 0.322, 0.05, 0.036)
    assert frq_qdsb < ((58, 0.322, 0.05, 0.036))
    assert frq_qdsb > FrequencyQuadrupleSideBand(57, 0.322, 0.04, 0.01)
    assert frq_qdsb > None
    assert (frq_qdsb < None) is False

    assert 57 != frq_qdsb
    assert 57.372 == frq_qdsb
    assert 56.646 == frq_qdsb
    assert 56.71 == frq_qdsb

    assert frq_qdsb != FrequencyQuadrupleSideBand(57, 0.322, 0.1, 0.040)

    frq_qdsb = None
    assert FrequencyQuadrupleSideBand(57, 0.322, 0.05, 0.036) != frq_qdsb
    assert frq_qdsb < FrequencyQuadrupleSideBand(57, 0.322, 0.05, 0.04)


def test_frequency_quadruple_side_band_channel_distances():
    """Test the frequency quadruple side band object: get the distance between two bands."""
    frq_qdsb = FrequencyQuadrupleSideBand(57, 0.322, 0.05, 0.036)
    mydist = frq_qdsb.distance([57, 0.322, 0.05, 0.036])

    frq_dict = {"central": 57, "side": 0.322, "sideside": 0.05,
                "bandwidth": 0.036, "unit": "GHz"}
    mydist = frq_qdsb.distance(frq_dict)
    assert mydist == np.inf

    mydist = frq_qdsb.distance(57.372)
    assert mydist == 0.0

    mydist = frq_qdsb.distance(FrequencyQuadrupleSideBand(57, 0.322, 0.05, 0.036))
    assert mydist == 0.0

    mydist = frq_qdsb.distance(57.38)
    np.testing.assert_almost_equal(mydist, 0.008)

    mydist = frq_qdsb.distance(57)
    assert mydist == np.inf

    mydist = frq_qdsb.distance((57, 0.322, 0.05, 0.018))
    assert mydist == np.inf


def test_frequency_quadruple_side_band_channel_containment():
    """Test the frequency quadruple side band object: check if one band contains another."""
    frq_qdsb = FrequencyQuadrupleSideBand(57, 0.322, 0.05, 0.036)

    assert 57 not in frq_qdsb
    assert 57.373 in frq_qdsb

    with pytest.raises(NotImplementedError):
        assert frq_qdsb in FrequencyQuadrupleSideBand(57, 0.322, 0.05, 0.05)

    frq_qdsb = None
    assert (frq_qdsb in FrequencyQuadrupleSideBand(57, 0.322, 0.05, 0.05)) is False

    assert "57" not in FrequencyQuadrupleSideBand(57, 0.322, 0.05, 0.05)


def test_frequency_double_side_band_class_method_convert():
    """Test the frequency double side band object: test the class method convert."""
    frq_dsb = FrequencyDoubleSideBand(183, 7, 2)

    res = frq_dsb.convert(185)
    assert res == 185

    res = frq_dsb.convert({"central": 185, "side": 7, "bandwidth": 2})
    assert res == FrequencyDoubleSideBand(185, 7, 2)


def test_frequency_double_side_band_channel_str():
    """Test the frequency double side band object: test the band description."""
    frq_dsb1 = FrequencyDoubleSideBand(183, 7, 2)
    frq_dsb2 = FrequencyDoubleSideBand(183000, 7000, 2000, "MHz")

    assert str(frq_dsb1) == "central=183 GHz ±7 width=2 GHz"
    assert str(frq_dsb2) == "central=183000 MHz ±7000 width=2000 MHz"


def test_frequency_double_side_band_channel_equality():
    """Test the frequency double side band object: check if two bands are 'equal'."""
    frq_dsb = FrequencyDoubleSideBand(183, 7, 2)
    assert frq_dsb is not None
    assert 183 != frq_dsb
    assert 190 == frq_dsb
    assert 176 == frq_dsb
    assert 175.5 == frq_dsb

    assert frq_dsb != FrequencyDoubleSideBand(183, 6.5, 3)

    frq_dsb = None
    assert FrequencyDoubleSideBand(183, 7, 2) != frq_dsb

    assert frq_dsb < FrequencyDoubleSideBand(183, 7, 2)

    assert FrequencyDoubleSideBand(182, 7, 2) < FrequencyDoubleSideBand(183, 7, 2)
    assert FrequencyDoubleSideBand(184, 7, 2) > FrequencyDoubleSideBand(183, 7, 2)


def test_frequency_double_side_band_channel_distances():
    """Test the frequency double side band object: get the distance between two bands."""
    frq_dsb = FrequencyDoubleSideBand(183, 7, 2)
    mydist = frq_dsb.distance(175.5)
    assert mydist == 0.5

    mydist = frq_dsb.distance(190.5)
    assert mydist == 0.5

    np.testing.assert_almost_equal(frq_dsb.distance(175.6), 0.4)
    np.testing.assert_almost_equal(frq_dsb.distance(190.1), 0.1)

    mydist = frq_dsb.distance(185)
    assert mydist == np.inf

    mydist = frq_dsb.distance((183, 7.0, 2))
    assert mydist == 0

    mydist = frq_dsb.distance((183, 7.0, 1))
    assert mydist == 0

    mydist = frq_dsb.distance(FrequencyDoubleSideBand(183, 7.0, 2))
    assert mydist == 0


def test_frequency_double_side_band_channel_containment():
    """Test the frequency double side band object: check if one band contains another."""
    frq_range = FrequencyDoubleSideBand(183, 7, 2)

    assert 175.5 in frq_range
    assert frq_range in FrequencyDoubleSideBand(183, 6.5, 3)
    assert frq_range not in FrequencyDoubleSideBand(183, 4, 2)

    with pytest.raises(NotImplementedError):
        assert frq_range in FrequencyDoubleSideBand(183, 6.5, 3, "MHz")

    frq_range = None
    assert (frq_range in FrequencyDoubleSideBand(183, 3, 2)) is False

    assert "183" not in FrequencyDoubleSideBand(183, 3, 2)


def test_frequency_range_class_method_convert():
    """Test the frequency range object: test the class method convert."""
    frq_range = FrequencyRange(89, 2)

    res = frq_range.convert(89)
    assert res == 89

    res = frq_range.convert({"central": 89, "bandwidth": 2})
    assert res == FrequencyRange(89, 2)


def test_frequency_range_class_method_str():
    """Test the frequency range object: test the band description."""
    frq_range1 = FrequencyRange(89, 2)
    frq_range2 = FrequencyRange(89000, 2000, "MHz")

    assert str(frq_range1) == "central=89 GHz width=2 GHz"
    assert str(frq_range2) == "central=89000 MHz width=2000 MHz"


def test_frequency_range_channel_equality():
    """Test the frequency range object: check if two bands are 'equal'."""
    frqr = FrequencyRange(2, 1)
    assert frqr is not None
    assert 1.7 == frqr
    assert 1.2 != frqr
    assert frqr == (2, 1)

    assert frqr == (2, 1, "GHz")


def test_frequency_range_channel_containment():
    """Test the frequency range object: channel containment."""
    frqr = FrequencyRange(2, 1)
    assert 1.7 in frqr
    assert 2.8 not in frqr

    with pytest.raises(NotImplementedError):
        assert frqr in FrequencyRange(89, 2, "MHz")

    frqr = None
    assert (frqr in FrequencyRange(89, 2)) is False

    assert "89" not in FrequencyRange(89, 2)


def test_frequency_range_channel_distances():
    """Test the frequency range object: derive distances between bands."""
    frqr = FrequencyRange(190.0, 2)

    mydist = frqr.distance(FrequencyRange(190, 2))
    assert mydist == 0
    mydist = frqr.distance(FrequencyRange(189.5, 2))
    assert mydist == np.inf
    mydist = frqr.distance(189.5)
    assert mydist == 0.5
    mydist = frqr.distance(188.0)
    assert mydist == np.inf


def test_wavelength_range():
    """Test the wavelength range object."""
    wr = WavelengthRange(1, 2, 3)
    assert 1.2 == wr
    assert .9 != wr
    assert wr == (1, 2, 3)
    assert wr == (1, 2, 3, "µm")

    # Check containement
    assert 1.2 in wr
    assert .9 not in wr
    assert WavelengthRange(1, 2, 3) in wr
    assert WavelengthRange(1.1, 2.2, 3.3) not in wr
    assert WavelengthRange(1.2, 2, 2.8) in wr
    assert WavelengthRange(10, 20, 30) not in wr
    assert "bla" not in wr
    assert None not in wr
    wr2 = WavelengthRange(1, 2, 3, "µm")
    assert wr2 in wr
    wr2 = WavelengthRange(1, 2, 3, "nm")
    with pytest.raises(NotImplementedError):
        wr2 in wr  # noqa

    # Check __str__
    assert str(wr) == "2 µm (1-3 µm)"
    assert str(wr2) == "2 nm (1-3 nm)"

    wr = WavelengthRange(10.5, 11.5, 12.5)
    np.testing.assert_almost_equal(wr.distance(11.1), 0.4)


def test_wavelength_range_cf_roundtrip():
    """Test the wavelength range object roundtrip to cf."""
    wr = WavelengthRange(1, 2, 3)

    assert WavelengthRange.from_cf(wr.to_cf()) == wr
    assert WavelengthRange.from_cf([str(item) for item in wr]) == wr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2017, 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Helpers for reading hdf5-based files."""

import logging
import os

import dask.array as da
import dask.config as dc
import h5py
import numpy as np
import xarray as xr
from dask.array.core import normalize_chunks
from dask.base import tokenize

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.remote import open_file_or_filename
from satpy.readers.core.utils import np2str

LOG = logging.getLogger(__name__)


class HDF5FileHandler(BaseFileHandler):
    """Small class for inspecting a HDF5 file and retrieve its metadata/header data."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize file handler."""
        super(HDF5FileHandler, self).__init__(
            filename, filename_info, filetype_info)
        self.file_content = {}
        self._attrs_cache = {}

        try:
            f_obj = open_file_or_filename(self.filename)
            file_handle = h5py.File(f_obj, "r")
        except IOError:
            LOG.exception(
                "Failed reading file %s. Possibly corrupted file", self.filename)
            raise

        file_handle.visititems(self.collect_metadata)
        self._collect_attrs("", file_handle.attrs)
        file_handle.close()

    def _collect_attrs(self, name, attrs):
        attrs_cache = self._attrs_cache.setdefault(name, {})
        for key, value in attrs.items():
            value = np.squeeze(value)
            fc_key = "{}/attr/{}".format(name, key)
            try:
                value = np2str(value)
            except ValueError:
                # use the original value
                pass
            except AttributeError:
                # A HDF5 reference ?
                value = self.get_reference(name, key)
                if value is None:
                    LOG.warning("Value cannot be converted - skip setting attribute %s", fc_key)
                    continue
            self.file_content[fc_key] = attrs_cache[key] = value

    def get_reference(self, name, key):
        """Get reference."""
        f_obj = open_file_or_filename(self.filename)
        with h5py.File(f_obj, "r") as hf:
            return self._get_reference(hf, hf[name].attrs[key])

    def _get_reference(self, hf, ref):
        try:
            return [self._get_reference(hf, elt) for elt in ref]
        except TypeError:
            if isinstance(ref, h5py.h5r.Reference):
                ref_name = h5py.h5r.get_name(ref, hf.id)
                return hf[ref_name][()]

    def collect_metadata(self, name, obj):
        """Collect metadata."""
        if isinstance(obj, h5py.Dataset):
            self.file_content[name] = obj
            self.file_content[name + "/dtype"] = obj.dtype
            self.file_content[name + "/shape"] = obj.shape
        self._collect_attrs(name, obj.attrs)

    def __getitem__(self, key):
        """Get item for given key."""
        val = self.file_content[key]
        if isinstance(val, h5py.Dataset):
            # these datasets are closed and inaccessible when the file is closed, need to reopen
            f_obj = open_file_or_filename(self.filename)
            dset = h5py.File(f_obj, "r")[key]
            dset_data = from_h5_array(dset)
            attrs = self._attrs_cache.get(key, dset.attrs)
            if dset.ndim == 2:
                return xr.DataArray(dset_data, dims=["y", "x"], attrs=attrs)
            return xr.DataArray(dset_data, attrs=attrs)

        return val

    def __contains__(self, item):
        """Get item from file content."""
        return item in self.file_content

    def get(self, item, default=None):
        """Get item."""
        if item in self:
            return self[item]
        else:
            return default


def from_h5_array(h5dset):
    """Create a dask array from an h5py dataset, ensuring uniqueness of the dask array name."""
    chunk_size = dc.get("array.chunk-size")

    chunks = normalize_chunks(chunk_size, dtype=h5dset.dtype, previous_chunks=h5dset.chunks, shape=h5dset.shape)
    name = h5dset.name + "-" + tokenize(os.fspath(h5dset.file.filename), h5dset.name, chunks)

    dset_data = da.from_array(h5dset, chunks=chunks, name=name)
    return dset_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Setup and configuration for all reader tests."""

from ._modis_fixtures import (  # noqa: F401, I001
    modis_l1b_imapp_1000m_file,
    modis_l1b_imapp_geo_file,
    modis_l1b_nasa_1km_mod03_files,
    modis_l1b_nasa_mod02hkm_file,
    modis_l1b_nasa_mod02qkm_file,
    modis_l1b_nasa_mod03_file,
    modis_l1b_nasa_mod021km_file,
    modis_l2_imapp_mask_byte1_file,
    modis_l2_imapp_mask_byte1_geo_files,
    modis_l2_imapp_snowmask_file,
    modis_l2_imapp_snowmask_geo_files,
    modis_l2_nasa_mod06_file,
    modis_l2_nasa_mod35_file,
    modis_l2_nasa_mod35_mod03_files,
    modis_l2_nasa_mod99_file,
    modis_l3_nasa_mcd12q1_file,
    modis_l3_nasa_mod09_file,
    modis_l3_nasa_mod43_file,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2014-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""GOES HRIT format reader.

References:
      LRIT/HRIT Mission Specific Implementation, February 2012
      GVARRDL98.pdf
      05057_SPE_MSG_LRIT_HRI

"""

import datetime as dt
import logging

import dask.array as da
import numpy as np
import xarray as xr
from numpy.typing import ArrayLike

from satpy.readers.core._geos_area import get_area_definition, get_area_extent, get_geos_area_naming
from satpy.readers.core.eum import recarray2dict, time_cds_short
from satpy.readers.core.hrit import (
    HRITFileHandler,
    ancillary_text,
    annotation_header,
    base_hdr_map,
    image_data_function,
)


class CalibrationError(Exception):
    """Dummy error-class."""


logger = logging.getLogger("hrit_goes")

# Geometric constants [meters]
EQUATOR_RADIUS = 6378169.00
POLE_RADIUS = 6356583.80
ALTITUDE = 35785831.00

# goes implementation:
key_header = np.dtype([("key_number", "u1"),
                       ("seed", ">f8")])

segment_identification = np.dtype([("GP_SC_ID", ">i2"),
                                   ("spectral_channel_id", ">i1"),
                                   ("segment_sequence_number", ">u2"),
                                   ("planned_start_segment_number", ">u2"),
                                   ("planned_end_segment_number", ">u2"),
                                   ("data_field_representation", ">i1")])

image_segment_line_quality = np.dtype([("line_number_in_grid", ">i4"),
                                       ("line_mean_acquisition",
                                        [("days", ">u2"),
                                         ("milliseconds", ">u4")]),
                                       ("line_validity", "u1"),
                                       ("line_radiometric_quality", "u1"),
                                       ("line_geometric_quality", "u1")])

goms_variable_length_headers = {
    image_segment_line_quality: "image_segment_line_quality"}

goms_text_headers = {image_data_function: "image_data_function",
                     annotation_header: "annotation_header",
                     ancillary_text: "ancillary_text"}

goes_hdr_map = base_hdr_map.copy()
goes_hdr_map.update({7: key_header,
                     128: segment_identification,
                     129: image_segment_line_quality
                     })


orbit_coef = np.dtype([("StartTime", time_cds_short),
                       ("EndTime", time_cds_short),
                       ("X", ">f8", (8, )),
                       ("Y", ">f8", (8, )),
                       ("Z", ">f8", (8, )),
                       ("VX", ">f8", (8, )),
                       ("VY", ">f8", (8, )),
                       ("VZ", ">f8", (8, ))])

attitude_coef = np.dtype([("StartTime", time_cds_short),
                          ("EndTime", time_cds_short),
                          ("XofSpinAxis", ">f8", (8, )),
                          ("YofSpinAxis", ">f8", (8, )),
                          ("ZofSpinAxis", ">f8", (8, ))])

cuc_time = np.dtype([("coarse", "u1", (4, )),
                     ("fine", "u1", (3, ))])


sgs_time = np.dtype([("century", "u1"),
                     ("year", "u1"),
                     ("doy1", "u1"),
                     ("doy_hours", "u1"),
                     ("hours_mins", "u1"),
                     ("mins_secs", "u1"),
                     ("secs_msecs", "u1"),
                     ("msecs", "u1")])


def make_sgs_time(sgs_time_array: ArrayLike) -> dt.datetime:
    """Make sgs time."""
    epoch_year = _epoch_year_from_sgs_time(sgs_time_array)
    doy_offset = _epoch_doy_offset_from_sgs_time(sgs_time_array)
    return epoch_year + doy_offset


def _epoch_year_from_sgs_time(sgs_time_array: ArrayLike) -> dt.datetime:
    century = sgs_time_array["century"].astype(np.int64)
    year = sgs_time_array["year"].astype(np.int64)
    year = ((century >> 4) * 1000 + (century & 15) * 100 + (year >> 4) * 10 + (year & 15))
    return dt.datetime(int(year), 1, 1)


def _epoch_doy_offset_from_sgs_time(sgs_time_array: ArrayLike) -> dt.timedelta:
    doy1 = sgs_time_array["doy1"].astype(np.int64)
    doy_hours = sgs_time_array["doy_hours"].astype(np.int64)
    hours_mins = sgs_time_array["hours_mins"].astype(np.int64)
    mins_secs = sgs_time_array["mins_secs"].astype(np.int64)
    secs_msecs = sgs_time_array["secs_msecs"].astype(np.int64)
    msecs = sgs_time_array["msecs"].astype(np.int64)

    doy = ((doy1 >> 4) * 100 + (doy1 & 15) * 10 + (doy_hours >> 4))
    hours = ((doy_hours & 15) * 10 + (hours_mins >> 4))
    mins = ((hours_mins & 15) * 10 + (mins_secs >> 4))
    secs = ((mins_secs & 15) * 10 + (secs_msecs >> 4))
    msecs = ((secs_msecs & 15) * 100 + (msecs >> 4) * 10 + (msecs & 15))
    return dt.timedelta(
        days=int(doy - 1),
        hours=int(hours),
        minutes=int(mins),
        seconds=int(secs),
        milliseconds=int(msecs)
    )


satellite_status = np.dtype([("TagType", "<u4"),
                             ("TagLength", "<u4"),
                             ("SatelliteID", "<u8"),
                             ("SatelliteName", "S256"),
                             ("NominalLongitude", "<f8"),
                             ("SatelliteCondition", "<u4"),
                             ("TimeOffset", "<f8")])

image_acquisition = np.dtype([("TagType", "<u4"),
                              ("TagLength", "<u4"),
                              ("Status", "<u4"),
                              ("StartDelay", "<i4"),
                              ("Cel", "<f8")])


gvar_float = ">i4"


def make_gvar_float(float_val):
    """Make gvar float."""
    sign = 1
    if float_val < 0:
        float_val = -float_val
        sign = -1

    exp = (float_val >> 24) - 64
    mant = float_val & ((1 << 24) - 1)
    if mant == 0:
        return 0.
    res = sign * mant * 2.0**(-24 + exp * 4)
    return res


prologue = np.dtype([
    # common generic header
    ("CommonHeaderVersion", "u1"),
    ("Junk1", "u1", 3),
    ("NominalSGSProductTime", time_cds_short),
    ("SGSProductQuality", "u1"),
    ("SGSProductCompleteness", "u1"),
    ("SGSProductTimeliness", "u1"),
    ("SGSProcessingInstanceId", "u1"),
    ("BaseAlgorithmVersion", "S1", 16),
    ("ProductAlgorithmVersion", "S1", 16),
    # product header
    ("ImageProductHeaderVersion", "u1"),
    ("Junk2", "u1", 3),
    ("ImageProductHeaderLength", ">u4"),
    ("ImageProductVersion", "u1"),
    # first block-0
    ("SatelliteID", "u1"),
    ("SPSID", "u1"),
    ("IScan", "u1", 4),
    ("IDSub", "u1", 16),
    ("TCurr", sgs_time),
    ("TCHED", sgs_time),
    ("TCTRL", sgs_time),
    ("TLHED", sgs_time),
    ("TLTRL", sgs_time),
    ("TIPFS", sgs_time),
    ("TINFS", sgs_time),
    ("TISPC", sgs_time),
    ("TIECL", sgs_time),
    ("TIBBC", sgs_time),
    ("TISTR", sgs_time),
    ("TLRAN", sgs_time),
    ("TIIRT", sgs_time),
    ("TIVIT", sgs_time),
    ("TCLMT", sgs_time),
    ("TIONA", sgs_time),
    ("RelativeScanCount", ">u2"),
    ("AbsoluteScanCount", ">u2"),
    ("NorthernmostScanLine", ">u2"),
    ("WesternmostPixel", ">u2"),
    ("EasternmostPixel", ">u2"),
    ("NorthernmostFrameLine", ">u2"),
    ("SouthernmostFrameLine", ">u2"),
    ("0Pixel", ">u2"),
    ("0ScanLine", ">u2"),
    ("0Scan", ">u2"),
    ("SubSatScan", ">u2"),
    ("SubSatPixel", ">u2"),
    ("SubSatLatitude", gvar_float),
    ("SubSatLongitude", gvar_float),
    ("Junk4", "u1", 96),  # move to "word" 295
    ("IMCIdentifier", "S4"),
    ("Zeros", "u1", 12),
    ("ReferenceLongitude", gvar_float),
    ("ReferenceDistance",  gvar_float),
    ("ReferenceLatitude",  gvar_float)
])


class HRITGOESPrologueFileHandler(HRITFileHandler):
    """GOES HRIT format reader."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the reader."""
        super(HRITGOESPrologueFileHandler, self).__init__(filename, filename_info,
                                                          filetype_info,
                                                          (goes_hdr_map,
                                                           goms_variable_length_headers,
                                                           goms_text_headers))
        self.prologue = {}
        self.read_prologue()

    def read_prologue(self):
        """Read the prologue metadata."""
        with open(self.filename, "rb") as fp_:
            fp_.seek(self.mda["total_header_length"])
            data = np.fromfile(fp_, dtype=prologue, count=1)
            self.prologue.update(recarray2dict(data))

        self.process_prologue()

    def process_prologue(self):
        """Reprocess prologue to correct types."""
        for key in ["TCurr", "TCHED", "TCTRL", "TLHED", "TLTRL", "TIPFS",
                    "TINFS", "TISPC", "TIECL", "TIBBC", "TISTR", "TLRAN",
                    "TIIRT", "TIVIT", "TCLMT", "TIONA"]:
            try:
                self.prologue[key] = make_sgs_time(self.prologue[key])
            except ValueError:
                self.prologue.pop(key, None)
                logger.debug("Invalid data for %s", key)

        for key in ["SubSatLatitude", "SubSatLongitude", "ReferenceLongitude",
                    "ReferenceDistance", "ReferenceLatitude"]:
            self.prologue[key] = make_gvar_float(self.prologue[key])


radiometric_processing = np.dtype([("TagType", "<u4"),
                                   ("TagLength", "<u4"),
                                   ("RPSummary",
                                    [("Impulse", "<u4"),
                                     ("IsStrNoiseCorrection", "<u4"),
                                        ("IsOptic", "<u4"),
                                        ("IsBrightnessAligment", "<u4")]),
                                   ("OpticCorrection",
                                    [("Degree", "<i4"),
                                     ("A", "<f8", (16, ))]),
                                   ("RPQuality",
                                    [("EffDinRange", "<f8"),
                                     ("EathDarkening", "<f8"),
                                        ("Zone", "<f8"),
                                        ("Impulse", "<f8"),
                                        ("Group", "<f8"),
                                        ("DefectCount", "<u4"),
                                        ("DefectProcent", "<f8"),
                                        ("S_Noise_DT_Preflight", "<f8"),
                                        ("S_Noise_DT_Bort", "<f8"),
                                        ("S_Noise_DT_Video", "<f8"),
                                        ("S_Noise_DT_1_5", "<f8"),
                                        ("CalibrStability", "<f8"),
                                        ("TemnSKO", "<f8", (2, )),
                                        ("StructSKO", "<f8", (2, )),
                                        ("Struct_1_5", "<f8"),
                                        ("Zone_1_ 5", "<f8"),
                                        ("RadDif", "<f8")])])

geometric_processing = np.dtype([("TagType", "<u4"),
                                 ("TagLength", "<u4"),
                                 ("TGeomNormInfo",
                                  [("IsExist", "<u4"),
                                   ("IsNorm", "<u4"),
                                      ("SubLon", "<f8"),
                                      ("TypeProjection", "<u4"),
                                      ("PixInfo", "<f8", (4, ))]),
                                 ("SatInfo",
                                  [("TISO",
                                    [("T0", "<f8"),
                                     ("dT", "<f8"),
                                        ("ASb", "<f8"),
                                        ("Evsk", "<f8", (3, 3, 4)),
                                        ("ARx", "<f8", (4, )),
                                        ("ARy", "<f8", (4, )),
                                        ("ARz", "<f8", (4, )),
                                        ("AVx", "<f8", (4, )),
                                        ("AVy", "<f8", (4, )),
                                        ("AVz", "<f8", (4, ))]),
                                      ("Type", "<i4")]),
                                 ("TimeProcessing", "<f8"),
                                 ("ApriorAccuracy", "<f8"),
                                 ("RelativeAccuracy", "<f8", (2, ))])


C1 = 1.19104273e-5
C2 = 1.43877523

SPACECRAFTS = {
    # these are GP_SC_ID
    18007: "GOES-7",
    18008: "GOES-8",
    18009: "GOES-9",
    18010: "GOES-10",
    18011: "GOES-11",
    18012: "GOES-12",
    18013: "GOES-13",
    18014: "GOES-14",
    18015: "GOES-15",
    # these are in block-0
    7: "GOES-7",
    8: "GOES-8",
    9: "GOES-9",
    10: "GOES-10",
    11: "GOES-11",
    12: "GOES-12",
    13: "GOES-13",
    14: "GOES-14",
    15: "GOES-15"}

SENSOR_NAME = "goes_imager"


class HRITGOESFileHandler(HRITFileHandler):
    """GOES HRIT format reader."""

    def __init__(self, filename, filename_info, filetype_info,
                 prologue):
        """Initialize the reader."""
        super(HRITGOESFileHandler, self).__init__(filename, filename_info,
                                                  filetype_info,
                                                  (goes_hdr_map,
                                                   goms_variable_length_headers,
                                                   goms_text_headers))
        self.prologue = prologue.prologue
        self.chid = self.mda["spectral_channel_id"]

        sublon = self.prologue["SubSatLongitude"]
        self.mda["projection_parameters"]["SSP_longitude"] = sublon

        satellite_id = self.prologue["SatelliteID"]
        self.platform_name = SPACECRAFTS[satellite_id]

    def get_dataset(self, key, info):
        """Get the data  from the files."""
        logger.debug("Getting raw data")
        res = super(HRITGOESFileHandler, self).get_dataset(key, info)

        self.mda["calibration_parameters"] = self._get_calibration_params()

        res = self.calibrate(res, key["calibration"])
        new_attrs = info.copy()
        new_attrs.update(res.attrs)
        res.attrs = new_attrs
        res.attrs["platform_name"] = self.platform_name
        res.attrs["sensor"] = SENSOR_NAME
        res.attrs["orbital_parameters"] = {"projection_longitude": self.mda["projection_parameters"]["SSP_longitude"],
                                           "projection_latitude": 0.0,
                                           "projection_altitude": ALTITUDE}
        return res

    def _get_calibration_params(self):
        """Get the calibration parameters from the metadata."""
        params = {}
        idx_table = []
        val_table = []
        for elt in self.mda["image_data_function"].split(b"\r\n"):
            try:
                key, val = elt.split(b":=")
                try:
                    idx_table.append(int(key))
                    val_table.append(float(val))
                except ValueError:
                    params[key] = val
            except ValueError:
                pass
        params["indices"] = np.array(idx_table)
        params["values"] = np.array(val_table, dtype=np.float32)
        return params

    def calibrate(self, data, calibration):
        """Calibrate the data."""
        logger.debug("Calibration")
        tic = dt.datetime.now()
        if calibration == "counts":
            return data
        if calibration == "reflectance":
            res = self._calibrate(data)
        elif calibration == "brightness_temperature":
            res = self._calibrate(data)
        else:
            raise NotImplementedError("Don't know how to calibrate to " +
                                      str(calibration))

        logger.debug("Calibration time " + str(dt.datetime.now() - tic))
        return res

    def _calibrate(self, data):
        """Calibrate *data*."""
        idx = self.mda["calibration_parameters"]["indices"]
        val = self.mda["calibration_parameters"]["values"]
        data.data = da.where(data.data == 0, np.nan, data.data)
        ddata = data.data.map_blocks(np.interp, idx, val, dtype=val.dtype)
        res = xr.DataArray(ddata,
                           dims=data.dims, attrs=data.attrs,
                           coords=data.coords)
        res = res.clip(min=0)
        units = {b"percent": "%", b"degree Kelvin": "K"}
        unit = self.mda["calibration_parameters"][b"_UNIT"]
        res.attrs["units"] = units.get(unit, unit)
        return res

    def get_area_def(self, dataset_id):
        """Get the area definition of the band."""
        proj_dict = self._get_proj_dict(dataset_id)
        area_extent = get_area_extent(proj_dict)
        area = get_area_definition(proj_dict, area_extent)
        self.area = area
        return area

    def _get_proj_dict(self, dataset_id):
        loff = np.float32(self.mda["loff"])
        nlines = np.int32(self.mda["number_of_lines"])
        loff = nlines - loff
        name_dict = get_geos_area_naming({
            "platform_name": self.platform_name,
            "instrument_name": SENSOR_NAME,
            # Partial scans are padded to full disk
            "service_name": "FD",
            "service_desc": "Full Disk",
            "resolution": dataset_id["resolution"]
        })
        return {
            "a": EQUATOR_RADIUS,
            "b": POLE_RADIUS,
            "ssp_lon": float(self.prologue["SubSatLongitude"]),
            "h": ALTITUDE,
            "proj": "geos",
            "units": "m",
            "a_name": name_dict["area_id"],
            "a_desc": name_dict["description"],
            "p_id": "",
            "nlines": nlines,
            "ncols": np.int32(self.mda["number_of_columns"]),
            "cfac": np.int32(self.mda["cfac"]),
            "lfac": np.int32(self.mda["lfac"]),
            "coff": np.float32(self.mda["coff"]),
            "loff": loff,
            "scandir": "N2S"
        }



"""Tests for the H-SAF H5 reader."""

import datetime as dt
import os

import h5py
import numpy as np
import pytest

from satpy import Scene
from satpy.area import get_area_def

# real shape is 916, 1902
SHAPE_SC = (916, 1902)
SHAPE_SC_COLORMAP = (256, 3)
AREA_X_OFFSET = 1211
AREA_Y_OFFSET = 62


@pytest.fixture(scope="session")
def sc_h5_file(tmp_path_factory):
    """Create a fake HSAF SC HDF5 file."""
    filename = tmp_path_factory.mktemp("data") / "h10_20221115_day_merged.H5"
    h5f = h5py.File(filename, mode="w")
    h5f.create_dataset("SC", SHAPE_SC, dtype=np.uint8)
    h5f.create_dataset("colormap", SHAPE_SC_COLORMAP, dtype=np.uint8)
    return str(filename)


def _get_scene_with_loaded_sc_datasets(filename):
    """Return a scene with SC and SC_pal loaded."""
    loaded_scene = Scene(filenames=[filename], reader="hsaf_h5")
    loaded_scene.load(["SC", "SC_pal"])
    return loaded_scene


def test_hsaf_sc_dataset(sc_h5_file):
    """Test the H-SAF SC dataset."""
    loaded_scene = _get_scene_with_loaded_sc_datasets(sc_h5_file)
    assert loaded_scene["SC"].shape == SHAPE_SC


def test_hsaf_sc_colormap_dataset(sc_h5_file):
    """Test the H-SAF SC_pal dataset."""
    loaded_scene = _get_scene_with_loaded_sc_datasets(sc_h5_file)
    assert loaded_scene["SC_pal"].shape == SHAPE_SC_COLORMAP


def test_hsaf_sc_datetime(sc_h5_file):
    """Test the H-SAF reference time."""
    loaded_scene = _get_scene_with_loaded_sc_datasets(sc_h5_file)
    fname = os.path.basename(sc_h5_file)
    dtstr = fname.split("_")[1]
    obs_time = dt.datetime.strptime(dtstr, "%Y%m%d")
    assert loaded_scene["SC"].attrs["data_time"] == obs_time


def test_hsaf_sc_areadef(sc_h5_file):
    """Test the H-SAF SC area definition."""
    loaded_scene = _get_scene_with_loaded_sc_datasets(sc_h5_file)
    fd_def = get_area_def("msg_seviri_fes_3km")
    hsaf_def = fd_def[AREA_Y_OFFSET:AREA_Y_OFFSET+SHAPE_SC[0], AREA_X_OFFSET:AREA_X_OFFSET+SHAPE_SC[1]]
    assert loaded_scene["SC"].area == hsaf_def



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.clavrx module."""

import os
import unittest
from unittest import mock

import numpy as np
import pytest
import xarray as xr
from pyresample.geometry import AreaDefinition

from satpy.tests.reader_tests.test_hdf4_utils import FakeHDF4FileHandler

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (10, 300)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LAT_DATA = np.repeat([DEFAULT_LAT_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_LON_DATA = np.linspace(5, 45, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LON_DATA = np.repeat([DEFAULT_LON_DATA], DEFAULT_FILE_SHAPE[0], axis=0)

class FakeHDF4FileHandlerGeo(FakeHDF4FileHandler):
    """Swap-in HDF4 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        file_content = {
            "/attr/platform": "HIM8",
            "/attr/sensor": "AHI",
            # this is a Level 2 file that came from a L1B file
            "/attr/L1B": "clavrx_H08_20180806_1800",
        }

        file_content["longitude"] = xr.DataArray(
            DEFAULT_LON_DATA,
            dims=("y", "x"),
            attrs={
                "_FillValue": np.nan,
                "scale_factor": 1.,
                "add_offset": 0.,
                "standard_name": "longitude",
            })
        file_content["longitude/shape"] = DEFAULT_FILE_SHAPE

        file_content["latitude"] = xr.DataArray(
            DEFAULT_LAT_DATA,
            dims=("y", "x"),
            attrs={
                "_FillValue": np.nan,
                "scale_factor": 1.,
                "add_offset": 0.,
                "standard_name": "latitude",
            })
        file_content["latitude/shape"] = DEFAULT_FILE_SHAPE

        file_content["refl_1_38um_nom"] = xr.DataArray(
            DEFAULT_FILE_DATA.astype(np.float32),
            dims=("y", "x"),
            attrs={
                "SCALED": 1,
                "add_offset": 59.0,
                "scale_factor": 0.0018616290763020515,
                "units": "%",
                "_FillValue": -32768,
                "valid_range": [-32767, 32767],
                "actual_range": [-2., 120.],
                "actual_missing": -999.0
            })
        file_content["refl_1_38um_nom/shape"] = DEFAULT_FILE_SHAPE

        # data with fill values
        file_content["variable2"] = xr.DataArray(
            DEFAULT_FILE_DATA.astype(np.float32),
            dims=("y", "x"),
            attrs={
                "_FillValue": -1,
                "scale_factor": 1.,
                "add_offset": 0.,
                "units": "1",
            })
        file_content["variable2/shape"] = DEFAULT_FILE_SHAPE
        file_content["variable2"] = file_content["variable2"].where(
            file_content["variable2"] % 2 != 0)

        # category
        file_content["variable3"] = xr.DataArray(
            DEFAULT_FILE_DATA.astype(np.byte),
            dims=("y", "x"),
            attrs={
                "SCALED": 0,
                "_FillValue": -128,
                "flag_meanings": "clear water supercooled mixed ice unknown",
                "flag_values": [0, 1, 2, 3, 4, 5],
                "units": "1",
            })
        file_content["variable3/shape"] = DEFAULT_FILE_SHAPE

        return file_content


class TestCLAVRXReaderGeo(unittest.TestCase):
    """Test CLAVR-X Reader with Geo files."""

    yaml_file = "clavrx.yaml"

    def setUp(self):
        """Wrap HDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.clavrx import CLAVRXHDF4FileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(CLAVRXHDF4FileHandler, "__bases__", (FakeHDF4FileHandlerGeo,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "clavrx_H08_20180806_1800.level2.hdf",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_no_nav_donor(self):
        """Test exception raised when no donor file is available."""
        import xarray as xr

        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        fake_fn = "clavrx_H08_20180806_1800.level2.hdf"
        with mock.patch("satpy.readers.clavrx.SDS", xr.DataArray):
            loadables = r.select_files_from_pathnames([fake_fn])
            r.create_filehandlers(loadables)
            l1b_base = fake_fn.split(".")[0]
            msg = f"Missing navigation donor {l1b_base}"
            with pytest.raises(IOError, match=msg):
                r.load(["refl_1_38um_nom", "variable2", "variable3"])

    def test_load_all_old_donor(self):
        """Test loading all test datasets with old donor."""
        import xarray as xr

        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.clavrx.SDS", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "clavrx_H08_20180806_1800.level2.hdf",
            ])
            r.create_filehandlers(loadables)
        with mock.patch("satpy.readers.clavrx.glob") as g, mock.patch("satpy.readers.clavrx.netCDF4.Dataset") as d:
            g.return_value = ["fake_donor.nc"]
            x = np.linspace(-0.1518, 0.1518, 300)
            y = np.linspace(0.1518, -0.1518, 10)
            proj = mock.Mock(
                semi_major_axis=6378.137,
                semi_minor_axis=6356.7523142,
                perspective_point_height=35791,
                longitude_of_projection_origin=140.7,
                sweep_angle_axis="y",
            )
            d.return_value = fake_donor = mock.MagicMock(
                variables={"Projection": proj, "x": x, "y": y},
            )
            fake_donor.__getitem__.side_effect = lambda key: fake_donor.variables[key]
            datasets = r.load(["refl_1_38um_nom", "variable2", "variable3"])
        assert len(datasets) == 3
        for v in datasets.values():
            assert "calibration" not in v.attrs
            assert v.attrs["units"] in ["1", "%"]
            assert isinstance(v.attrs["area"], AreaDefinition)
            if v.attrs.get("flag_values"):
                assert "_FillValue" in v.attrs
            else:
                assert "_FillValue" not in v.attrs
            if v.attrs["name"] == "refl_1_38um_nom":
                assert "valid_range" in v.attrs
                assert isinstance(v.attrs["valid_range"], list)
            else:
                assert "valid_range" not in v.attrs
            if "flag_values" in v.attrs:
                assert np.issubdtype(v.dtype, np.integer)
                assert v.attrs.get("flag_meanings") is not None

    def test_load_all_new_donor(self):
        """Test loading all test datasets with new donor."""
        import xarray as xr

        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.clavrx.SDS", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "clavrx_H08_20180806_1800.level2.hdf",
            ])
            r.create_filehandlers(loadables)
        with mock.patch("satpy.readers.clavrx.glob") as g, mock.patch("satpy.readers.clavrx.netCDF4.Dataset") as d:
            g.return_value = ["fake_donor.nc"]
            x = np.linspace(-0.1518, 0.1518, 300)
            y = np.linspace(0.1518, -0.1518, 10)
            proj = mock.Mock(
                semi_major_axis=6378137,
                semi_minor_axis=6356752.3142,
                perspective_point_height=35791000,
                longitude_of_projection_origin=140.7,
                sweep_angle_axis="y",
            )
            d.return_value = fake_donor = mock.MagicMock(
                variables={"goes_imager_projection": proj, "x": x, "y": y},
            )
            fake_donor.__getitem__.side_effect = lambda key: fake_donor.variables[key]
            datasets = r.load(["refl_1_38um_nom", "variable2", "variable3"])
        assert len(datasets) == 3
        for v in datasets.values():
            assert "calibration" not in v.attrs
            assert v.attrs["units"] in ["1", "%"]
            assert isinstance(v.attrs["area"], AreaDefinition)
            assert v.attrs["area"].is_geostationary is True
            assert v.attrs["platform_name"] == "himawari8"
            assert v.attrs["sensor"] == "ahi"
        assert datasets["variable3"].attrs.get("flag_meanings") is not None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2011-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Interface to VIIRS L1B format."""

import datetime as dt
import logging

import numpy as np

from satpy.readers.core.netcdf import NetCDF4FileHandler

LOG = logging.getLogger(__name__)


class VIIRSL1BFileHandler(NetCDF4FileHandler):
    """VIIRS L1B File Reader."""

    def _parse_datetime(self, datestr):
        """Parse datetime."""
        return dt.datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%S.000Z")

    @property
    def start_orbit_number(self):
        """Get start orbit number."""
        try:
            return int(self["/attr/orbit_number"])
        except KeyError:
            return int(self["/attr/OrbitNumber"])

    @property
    def end_orbit_number(self):
        """Get end orbit number."""
        try:
            return int(self["/attr/orbit_number"])
        except KeyError:
            return int(self["/attr/OrbitNumber"])

    @property
    def platform_name(self):
        """Get platform name."""
        try:
            res = self.get("/attr/platform",
                           self.filename_info["platform_shortname"])
        except KeyError:
            res = "Unknown"

        return {
            "JPSS-1": "NOAA-20",
            "NP": "Suomi-NPP",
            "J1": "NOAA-20",
            "J2": "NOAA-21",
            "JPSS-2": "NOAA-21",
        }.get(res, res)

    @property
    def sensor_name(self):
        """Get sensor name."""
        return self["/attr/instrument"].lower()

    def adjust_scaling_factors(self, factors, file_units, output_units):
        """Adjust scaling factors."""
        if factors is None or factors[0] is None:
            factors = [1, 0]
        if file_units == output_units:
            LOG.debug("File units and output units are the same (%s)", file_units)
            return factors
        factors = np.array(factors)

        if file_units == "W cm-2 sr-1" and output_units == "W m-2 sr-1":
            LOG.debug("Adjusting scaling factors to convert '%s' to '%s'", file_units, output_units)
            factors[::2] = np.where(factors[::2] != -999, factors[::2] * 10000.0, -999)
            factors[1::2] = np.where(factors[1::2] != -999, factors[1::2] * 10000.0, -999)
            return factors
        elif file_units == "1" and output_units == "%":
            LOG.debug("Adjusting scaling factors to convert '%s' to '%s'", file_units, output_units)
            factors[::2] = np.where(factors[::2] != -999, factors[::2] * 100.0, -999)
            factors[1::2] = np.where(factors[1::2] != -999, factors[1::2] * 100.0, -999)
            return factors
        else:
            return factors

    def get_shape(self, ds_id, ds_info):
        """Get shape."""
        var_path = self._dataset_name_to_var_path(ds_id["name"], ds_info)
        return self.get(var_path + "/shape", 1)

    @property
    def start_time(self):
        """Get start time."""
        return self._parse_datetime(self["/attr/time_coverage_start"])

    @property
    def end_time(self):
        """Get end time."""
        return self._parse_datetime(self["/attr/time_coverage_end"])

    def _get_dataset_file_units(self, dataset_id, ds_info, var_path):
        file_units = ds_info.get("file_units")
        if file_units is None:
            file_units = self.get(var_path + "/attr/units")
            # they were almost completely CF compliant...
            if file_units == "none":
                file_units = "1"

        if dataset_id.get("calibration") == "radiance" and ds_info["units"] == "W m-2 um-1 sr-1":
            rad_units_path = var_path + "/attr/radiance_units"
            if rad_units_path in self:
                if file_units is None:
                    file_units = self[var_path + "/attr/radiance_units"]
                if file_units == "Watts/meter^2/steradian/micrometer":
                    file_units = "W m-2 um-1 sr-1"
        elif ds_info.get("units") == "%" and file_units is None:
            # v1.1 and above of level 1 processing removed 'units' attribute
            # for all reflectance channels
            file_units = "1"

        return file_units

    def _get_dataset_valid_range(self, dataset_id, ds_info, var_path):
        if dataset_id.get("calibration") == "radiance" and ds_info["units"] == "W m-2 um-1 sr-1":
            rad_units_path = var_path + "/attr/radiance_units"
            if rad_units_path in self:
                # we are getting a reflectance band but we want the radiance values
                # special scaling parameters
                scale_factor = self[var_path + "/attr/radiance_scale_factor"]
                scale_offset = self[var_path + "/attr/radiance_add_offset"]
            else:
                # we are getting a btemp band but we want the radiance values
                # these are stored directly in the primary variable
                scale_factor = self[var_path + "/attr/scale_factor"]
                scale_offset = self[var_path + "/attr/add_offset"]
            valid_min = self[var_path + "/attr/valid_min"]
            valid_max = self[var_path + "/attr/valid_max"]
        elif ds_info.get("units") == "%":
            # normal reflectance
            valid_min = self[var_path + "/attr/valid_min"]
            valid_max = self[var_path + "/attr/valid_max"]
            scale_factor = self[var_path + "/attr/scale_factor"]
            scale_offset = self[var_path + "/attr/add_offset"]
        elif ds_info.get("units") == "K":
            # normal brightness temperature
            # use a special LUT to get the actual values
            lut_var_path = ds_info.get("lut", var_path + "_brightness_temperature_lut")
            # we get the BT values from a look up table using the scaled radiance integers
            valid_min = self[lut_var_path + "/attr/valid_min"]
            valid_max = self[lut_var_path + "/attr/valid_max"]
            scale_factor = scale_offset = None
        else:
            valid_min = self.get(var_path + "/attr/valid_min")
            valid_max = self.get(var_path + "/attr/valid_max")
            scale_factor = self.get(var_path + "/attr/scale_factor")
            scale_offset = self.get(var_path + "/attr/add_offset")

        return valid_min, valid_max, scale_factor, scale_offset

    def get_metadata(self, dataset_id, ds_info):
        """Get metadata."""
        var_path = self._dataset_name_to_var_path(dataset_id["name"], ds_info)
        shape = self.get_shape(dataset_id, ds_info)
        file_units = self._get_dataset_file_units(dataset_id, ds_info, var_path)

        # Get extra metadata
        if self._is_scan_based_array(shape):
            rows_per_scan = int(shape[0] / self["/dimension/number_of_scans"])
            ds_info.setdefault("rows_per_scan", rows_per_scan)

        i = getattr(self[var_path], "attrs", {})
        i.update(ds_info)
        i.update(dataset_id.to_dict())
        orb_param = {"start_direction": self["/attr/startDirection"],
                     "end_direction": self["/attr/endDirection"],
                     "start_orbit": self.start_orbit_number,
                     "end_orbit": self.end_orbit_number,}
        i.update({
            "shape": shape,
            "units": ds_info.get("units", file_units),
            "file_units": file_units,
            "platform_name": self.platform_name,
            "sensor": self.sensor_name,
            "day_night": self["/attr/DayNightFlag"],
            "orbital_parameters": orb_param,
            "start_orbit": self.start_orbit_number,
            "end_orbit": self.end_orbit_number,
        })
        i.update(dataset_id.to_dict())
        return i

    def _is_scan_based_array(self, shape):
        return "/dimension/number_of_scans" in self and isinstance(shape, tuple) and shape

    def get_dataset(self, dataset_id, ds_info):
        """Get dataset."""
        var_path = self._dataset_name_to_var_path(dataset_id["name"], ds_info)
        metadata = self.get_metadata(dataset_id, ds_info)

        valid_min, valid_max, scale_factor, scale_offset = self._get_dataset_valid_range(dataset_id, ds_info, var_path)
        if dataset_id.get("calibration") == "radiance" and ds_info["units"] == "W m-2 um-1 sr-1":
            data = self[var_path]
        elif ds_info.get("units") == "%":
            data = self[var_path]
        elif ds_info.get("units") == "K":
            # normal brightness temperature
            # use a special LUT to get the actual values
            lut_var_path = ds_info.get("lut", var_path + "_brightness_temperature_lut")
            data = self[var_path]
            # we get the BT values from a look up table using the scaled radiance integers
            index_arr = data.data.astype(int)
            coords = data.coords
            data.data = self[lut_var_path].data[index_arr.ravel()].reshape(data.shape)
            data = data.assign_coords(**coords)
        else:
            data = self[var_path]
        data.attrs.update(metadata)

        if valid_min is not None and valid_max is not None:
            data = data.where((data >= valid_min) & (data <= valid_max))
        if data.attrs.get("units") in ["%", "K", "1", "W m-2 um-1 sr-1"] and \
                "flag_meanings" in data.attrs:
            # flag meanings don't mean anything anymore for these variables
            # these aren't category products
            data.attrs.pop("flag_meanings", None)
            data.attrs.pop("flag_values", None)

        factors = (scale_factor, scale_offset)
        factors = self.adjust_scaling_factors(factors, metadata["file_units"], ds_info.get("units"))
        if factors[0] != 1 or factors[1] != 0:
            data *= factors[0]
            data += factors[1]

        # remove scale_factor and scale_offset
        data.attrs.pop("scale_factor", None)
        data.attrs.pop("add_offset", None)

        # rename dimensions to correspond to satpy's 'y' and 'x' standard
        if "number_of_lines" in data.dims:
            data = data.rename({"number_of_lines": "y", "number_of_pixels": "x"})
        return data

    def available_datasets(self, configured_datasets=None):
        """Generate dataset info and their availablity.

        See
        :meth:`satpy.readers.core.file_handlers.BaseFileHandler.available_datasets`
        for details.

        """
        for is_avail, ds_info in (configured_datasets or []):
            if is_avail is not None:
                # some other file handler said it has this dataset
                # we don't know any more information than the previous
                # file handler so let's yield early
                yield is_avail, ds_info
                continue
            ft_matches = self.file_type_matches(ds_info["file_type"])
            var_path = self._dataset_name_to_var_path(ds_info["name"], ds_info)
            is_in_file = var_path in self
            yield ft_matches and is_in_file, ds_info

            # Handle the _quality_flags dataset
            qf_name = ds_info["name"] + "_quality_flags"

            # Create a copy and override file_key if not explicitly provided
            qf_ds_info = ds_info.copy()
            qf_ds_info["name"] = qf_name
            qf_ds_info["file_key"] = f"observation_data/{qf_name}"
            qf_var_path = self._dataset_name_to_var_path(qf_name, qf_ds_info)

            if qf_var_path in self:
                qf_info = {
                    "name": qf_name,
                    "file_type": ds_info["file_type"],
                    "resolution": ds_info.get("resolution"),
                    "coordinates": ds_info.get("coordinates") or [],
                    "units": "1",
                    "standard_name": "quality_flag",
                }
                yield True, qf_info

    @staticmethod
    def _dataset_name_to_var_path(dataset_name: str, ds_info: dict) -> str:
        return ds_info.get("file_key", "observation_data/{}".format(dataset_name))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
# type: ignore

"""Interface to CALIOP L2 HDF4 cloud products."""

import datetime as dt
import logging
import os.path
import re

from pyhdf.SD import SD, SDC

from satpy.dataset import Dataset
from satpy.readers.core.file_handlers import BaseFileHandler

logger = logging.getLogger(__name__)


class HDF4BandReader(BaseFileHandler):
    """CALIOP v3 HDF4 reader."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialze file handler."""
        super(HDF4BandReader, self).__init__(filename,
                                             filename_info,
                                             filetype_info)
        self.lons = None
        self.lats = None
        self._start_time = None
        self._end_time = None

        self.get_filehandle()

        self._start_time = filename_info["start_time"]

        logger.debug("Retrieving end time from metadata array")
        self.get_end_time()

    def get_end_time(self):
        """Get observation end time from file metadata."""
        mda_dict = self.filehandle.attributes()
        core_mda = mda_dict["coremetadata"]
        end_time_str = self.parse_metadata_string(core_mda)
        self._end_time = dt.datetime.strptime(end_time_str, "%Y-%m-%dT%H:%M:%SZ")

    @staticmethod
    def parse_metadata_string(metadata_string):
        """Grab end time with regular expression."""
        regex = r"STOP_DATE.+?VALUE\s*=\s*\"(.+?)\""
        match = re.search(regex, metadata_string, re.DOTALL)
        end_time_str = match.group(1)
        return end_time_str

    def get_filehandle(self):
        """Get HDF4 filehandle."""
        if os.path.exists(self.filename):
            self.filehandle = SD(self.filename, SDC.READ)
            logger.debug("Loading dataset {}".format(self.filename))
        else:
            raise IOError("Path {} does not exist.".format(self.filename))

    def get_dataset(self, key, info):
        """Read data from file and return the corresponding projectables."""
        if key["name"] in ["longitude", "latitude"]:
            logger.debug("Reading coordinate arrays.")

            if self.lons is None or self.lats is None:
                self.lons, self.lats = self.get_lonlats()

            if key["name"] == "latitude":
                proj = Dataset(self.lats, id=key, **info)
            else:
                proj = Dataset(self.lons, id=key, **info)

        else:
            data = self.get_sds_variable(key["name"])
            proj = Dataset(data, id=key, **info)

        return proj

    def get_sds_variable(self, name):
        """Read variable from the HDF4 file."""
        sds_obj = self.filehandle.select(name)
        data = sds_obj.get()
        return data

    def get_lonlats(self):
        """Get longitude and latitude arrays from the file."""
        longitudes = self.get_sds_variable("Longitude")
        latitudes = self.get_sds_variable("Latitude")
        return longitudes, latitudes

    @property
    def start_time(self):
        """Get start time."""
        return self._start_time

    @property
    def end_time(self):
        """Get end time."""
        return self._end_time
